#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **set classes** (i.e., classes implementing set-like functionality,
typically by subclassing the builtin :class:`set` or :class:`frozenset`
container types or analogues thereof).
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.util.type.obj import objects
from betse.util.type.types import (
    type_check, ClassType, IterableTypes, MappingType,)
from functools import wraps

# ....................{ GLOBALS                           }....................
_FROZENSET_METACLASS = type(frozenset)
'''
Metaclass of the builtin "frozenset" container type.
'''

# ....................{ METACLASSES                       }....................
class FrozenSetSubclassableMeta(ABCMeta, _FROZENSET_METACLASS):
    '''
    Metaclass of the abstract :class:`FrozenSetSubclassable` base class and all
    concrete subclasses thereof.

    This metaclass dynamically redefines *all* container-creating methods of
    the :class:`frozenset` superclass of the :class:`FrozenSetSubclassable`
    class (e.g., :meth:`frozenset.__or__`) within the currently declared
    concrete subclass of that class. Specifically, this metaclass redefines
    each such method to return an instance of the currently declared concrete
    subclass of :class:`FrozenSetSubclassable` rather than of the
    :class:`frozenset` superclass, preserving sane semantics and caller
    expectations.

    Design
    ----------
    The :class:`FrozenSetSubclassable` class is merely a placeholder subclass
    of the :class:`frozenset` type whose metaclass is this metaclass. Ideally,
    the work performed by this metaclass would directly reside in the
    :class:`FrozenSetSubclassable` class instead, in which case this metaclass
    would have *no* demonstrable reason to exist. However, this work requires
    access to the concrete subclass of the :class:`FrozenSetSubclassable` class
    being currently declared. Since each such subclass is accessible *only*
    from within the metaclass of the :class:`FrozenSetSubclassable` class
    rather than within that class itself, this work necessarily resides in this
    metaclass.

    For general-purpose usability, this metaclass subclasses both:

    * The metaclass of the :class:`frozenset` type (typically, the root
      metaclass :class:`type`), avoiding conflicts between this metaclass and
      the :class:`FrozenSetSubclassable` class subclassing the
      :class:`frozenset` type.
    * The :class:`ABCMeta` metaclass, avoiding conflicts between this metaclass
      and concrete subclasses of the :class:`FrozenSetSubclassable` class which
      additionally subclass one or more other abstract base classes which
      themselves leverage the :class:`ABCMeta` metaclass. (Subtle dragons.)

    For these and similar reasons, metaclass usage in Python should typically
    be kept to a minimum. The :class:`FrozenSetSubclassable` class violates
    this maxim because it absolutely must; in all other cases, alternate
    solutions *not* leveraging metaclasses should be implemented instead.

    See Also
    ----------
    https://stackoverflow.com/a/804973/2809027
        StackOverflow answer mildly inspiring this class.
    '''

    # ..................{ CONSTRUCTORS                      }..................
    def __new__(
        metacls: ClassType,
        class_name: str,
        class_base_classes: IterableTypes,
        class_attrs: MappingType,
        **kwargs
    ) -> ClassType:
        '''
        Redefine all container-creating methods of the :class:`frozenset`
        superclass of the :class:`FrozenSetSubclassable` class (e.g.,
        :meth:`frozenset.__or__`) within the currently declared concrete
        subclass of that class identified by the passed parameters.
        '''

        # Tuple of the unqualified names of all container-creating methods
        # defined by the "frozenset" type, requiring redefinition in the
        # "FrozenSetSubclassable" subclass declared above.
        CREATION_METHOD_NAMES = (
            'copy',
            'difference',
            'intersection',
            'symmetric_difference',
            'union',
            '__and__',
            '__or__',
            '__rand__',
            '__ror__',
            '__rsub__',
            '__rxor__',
            '__sub__',
            '__xor__',
        )

        # Unsanitized "FrozenSetSubclassable" subclass.
        frozenset_subclass = super().__new__(
            metacls, class_name, class_base_classes, class_attrs)

        # For the name of each such method...
        for creation_method_name in CREATION_METHOD_NAMES:
            metacls._sanitize_creation_method(
                frozenset_subclass, creation_method_name)

        # Return this sanitized "FrozenSetSubclassable" subclass.
        return frozenset_subclass

    # ..................{ SANITIZERS                        }..................
    @staticmethod
    @type_check
    def _sanitize_creation_method(
        frozenset_subclass: ClassType, method_name: str) -> None:
        '''
        Redefine the container-creating method of the
        :class:`frozenset` superclass with the passed name in a cleverly
        automated manner circumventing all superclass issues.

        Design
        ----------
        This private static method is intended to be called only by the special
        static :meth:`__new__` method of this metaclass.

        This method is intentionally implemented as a discrete callable rather
        than inlined directly into the body of the :meth:`__new__` method, as
        the closure internally defined by this method expects the local
        ``frozenset_method`` variable captured by this closure to remain
        constant for the lifetime of that closure.

        Parameters
        ----------
        frozenset_subclass : ClassType
            :class:`FrozenSetSubclassable` subclass to redefine this method
            for.
        method_name : str
            Name of the container-creating method to be redefined.
        '''

        # Container-creating method with this name defined by "frozenset".
        frozenset_method = objects.get_callable(
            obj=frozenset, callable_name=method_name)

        # Closure sanitizing this method to return instances of the concrete
        # subclass inheriting from this class rather than of "frozenset",
        # wrapped in a manner propagating the name, docstring, and other
        # identifying metadata of the original method to this closure.
        @wraps(frozenset_method)
        def sanitized_creation_method(self, *args, **kwargs):

            # "frozenset" instance created by calling the superclass method
            # with all passed positional and keyword arguments.
            set_created = frozenset_method(self, *args, **kwargs)

            # Return either...
            return (
                # A new instance of this concrete subclass if the returned
                # object is in fact a "frozenset" instance.
                frozenset_subclass(set_created) if isinstance(
                    set_created, frozenset) else
                # This object as is otherwise. Technically, this should
                # probably never occur. Pragmatically, you know what they say
                # about every bad assumption we've ever made.
                set_created)

        # Override the superclass method with this closure.
        setattr(frozenset_subclass, method_name, sanitized_creation_method)

# ....................{ SUPERCLASSES                      }....................
class FrozenSetSubclassable(frozenset, metaclass=FrozenSetSubclassableMeta):
    '''
    Safely subclassable immutable set.

    Caveats
    ----------
    **Immutable set subclasses should always inherit from this base class.**
    Neither the builtin :class:`frozenset` container type nor the abstract
    :class:`collections.abc.Set` and :class:`collections.abc.Hashable` base
    classes should be inherited from.

    Subclasses should avoid declaring a metaclass or inheriting from another
    base class that declares a metaclass - excluding the standard
    :class:`ABCMeta` metaclass, which is compatible with this class by design.
    All other metaclasses should be considered incompatible. Violating this
    constraint typically raises the following runtime exception:

        TypeError: metaclass conflict: the metaclass of a derived class must be
        a (non-strict) subclass of the metaclasses of all its bases

    Subclasses must redefine the static ``__new__()`` method and *not* attempt
    to define the ``__init__()`` method. Immutable types are necessarily
    initialized at object creation time.

    By Python design, the ``__new__()`` method is static rather than a
    classmethod and hence *must* be manually redefined in *all* subclasses.
    This redefinition *must* internally call the superclass
    :func:`frozenset.__new__` method and return the result of doing so (i.e.,
    an instance of the desired subclass type). This redefinition should ideally
    (but *not* necessarily) share a similar signature as the
    :func:`frozenset.__new__` method and pass all passed parameters to that
    method as is. In positional order, these are:

    #. **The type of the current subclass.** Since the ``__new__()`` method is
       static and hence *not* bound to the type of the current subclass, this
       type *must* be manually passed as the first argument to this method.
    #. **The optional iterable defining the contents of this immutable set.**
       If unpassed, this set reduces to the empty set.

    The trivial implementation of the ``__new__()`` method is as follows:

        def __new__(cls, *args):
            return super().__new__(cls, *args)

    Versus :class:`frozenset`
    ----------
    The :class:`frozenset` type is *not* safely subclassable, due to
    unfortunate design decisions baked into the C-based implementations of
    *all* builtin container types. The core issue pertains to the objects
    returned by **container-creating methods** (i.e., methods declared by these
    types that create and return instances of the same types). In the case of
    :class:`frozenset`, these methods include:

    * Binary set operations (e.g., the set union operator `|`, internally
      implemented by the :meth:`frozenset.__or__` special method).
    * Binary set methods (e.g., the set union method :meth:`frozenset.union`).
    * Copy set methods (e.g., the set copying method :meth:`frozenset.copy`).

    The specifics of the unsuitability of :class:`frozenset` depend on the
    major version of Python in use. Specifically, under:

    * Python 2.x, container-creating methods in both the :class:`frozenset` and
      :class:`set` types correctly created instances of subclasses inheriting
      from these types but incorrectly failed to call the `__init__` methods of
      these subclasses. This is referred to as `issue #1721812`_.

    .. _issue #1721812:
        https://mail.python.org/pipermail/python-bugs-list/2007-May/038471.html

    * Python 3.x, container-creating methods in both the :class:`frozenset` and
      :class:`set` types incorrectly resolved `issue #1721812`_ by creating
      instances of the corresponding base types (e.g., :class:`frozenset` or
      :class:`set`) in subclasses inheriting from these types rather than
      instances of these subclasses.

    Technically, the latter issue may be resolved by manually redefining *all*
    container-creating methods in :class:`frozenset` subclasses. Pragmatically,
    there exist at least 17 such methods, rendering such redefinition
    effectively infeasible: ``__ror__``, ``difference_update``, ``__isub__``,
    ``symmetric_difference``, ``__rsub__``, ``__and__``, ``__rand__``,
    ``intersection``, ``difference``, ``__iand__``, ``union``, ``__ixor__``,
    ``symmetric_difference_update``, ``__or__``, ``copy``, ``__rxor__``,
    ``intersection_update``, ``__xor__``, ``__ior__``, and ``__sub__``.

    Versus ``Set`` and ``Hashable``
    ----------
    Technically, the abstract :class:`collections.abc.Set` base class defining
    the official immutable set API *is* safely subclassable, but only under the
    following stipulations:

    * The abstract :class:`collections.abc.Hashable` base class should
      typically also be subclassed. Failing to do so raises exceptions on
      attempting to add subclass instances to builtin container types expecting
      hashable objects (e.g., :class:`dict`, :class:`set`).
    * All public methods defined by the :class:`frozenset` type (e.g.,
      :meth:`frozenset.union`, :meth:`frozenset.issubset`) should typically
      also be defined, both for usability *and* duck typing purposes.

    Satisfying these stipulations requires defining the following 13 methods:
    ``__contains__``, ``__eq__``, ``__hash__``, ``__init__``, ``__iter__``,
    ``__len__``, ``difference``, ``intersection``, ``symetric_difference``,
    ``union``, ``copy``, ``issubset``, and ``issuperset``.

    While feasible, doing so is surprisingly less trivial than redefining the
    17 container-creating :class:`frozenset` methods listed above. The latter
    all share the exact same semantics and hence are trivially redefined with
    automation, as this class demonstrates. The former, however, share *no*
    common semantics and hence each require manual redefinition.

    Moreover, the typical implementation of a :class:`collections.abc.Set`
    subclass encapsulates an internal :class:`frozenset` instance variable.
    Since both approaches require the :class:`frozenset` class *and* since
    directly subclassing that class is simpler than indirectly encapsulating
    that class in :class:`collections.abc.Set` subclasses, this class elects to
    directly subclass the :class:`frozenset` type instead.
    '''

    # ..................{ CONSTRUCTORS                      }..................
    # The entirety of the logic for this class resides in our metaclass.
    def __new__(cls, *args):
        return super().__new__(cls, *args)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Global test configuration** (i.e., early-time configuration guaranteed to be
run by :mod:`pytest` *after* passed command-line arguments are parsed) for
this test suite.

:mod:`pytest` implicitly imports *all* functionality defined by this module
into *all* submodules of this subpackage.

See Also
--------
:mod:`conftest`
    Root test configuration applied before this configuration.
'''

# ....................{ IMPORTS                            }....................
import pytest

# ....................{ IMPORTS ~ fixture : manual         }....................
# Import fixtures required to be manually required by other fixtures and tests.

from betse.util.test.pytest.fixture.pytfixture import monkeypatch_session
from betse_test._fixture.tempdirer import betse_temp_dir
from betse_test._fixture.simconf.simconfer import (
    betse_sim_conf,
    betse_sim_conf_default,
    betse_sim_conf_compat,
)

# ....................{ IMPORTS ~ fixture : autouse        }....................
# Import fixtures automatically run at the start of the current test session,
# typically *NOT* manually required by specific tests, *AFTER* importing all
# non-autouse fixtures possibly required by these autouse fixtures above.

#FIXME: This fixture has been temporarily superceded by the
#betse_test.conftest.pytest_configure() and pytest_unconfigure() hooks until
#inevitably required yet again.
# from betse_test._fixture.initter import betse_init_session

# ....................{ HOOKS ~ configure                  }....................
def pytest_configure(config) -> None:
    '''
    Hook run immediately *after* parsing all command-line options and loading
    all third-party :mod:`pytest` plugins (including application-specific
    ``conftest`` scripts) but *before* performing test collection.

    Specifically, this hook (in no particular order):

    * Instantiates and initializes the application metadata singleton in a
      manner suitable for unit testing.

    See Also
    --------
    :func:`_init_app`
        Further details on application initialization.
    '''

    # Defer heavyweight imports.
    from betse_test._fixture import initter

    # Prepend a leading newline, which py.test curiously neglects to do itself.
    print('\n')

    # Initialize the application metadata singleton.
    initter.init_app()


def pytest_unconfigure(config) -> None:
    '''
    Hook run immediately *before* exiting the current test session.
    '''

    # Defer heavyweight imports.
    from betse.util.test.pytest import pytests
    from betse_test._fixture import initter

    # Deinitialize the application metadata singleton.
    initter.deinit_app()

    # Print this closure.
    pytests.output('Ego, ergo simulare.')

# ....................{ HOOKS ~ test                       }....................
def pytest_runtest_setup(item) -> None:
    '''
    Hook run immediately *before* running the passed test.

    Specifically:

    * If this is a **serial test** (i.e., test method bound to an instance of
      the :class:`betse.util.test.pytest.pytabc.SerialTestABC` superclass) for
      which a prior serial test in the same test class was recorded as failing
      by the :func:`pytest_runtest_makereport` hook, this hook marks this test
      as xfailing.

    Parameters
    ----------
    item : pytest.main.Item
        Metadata encapsulating this test callable (e.g., function, method).

    See Also
    --------
    https://pytest.org/latest/example/simple.html#incremental-testing-test-steps
        Official py.test code snippet inspiring this implementation.
    '''

    # Defer heavyweight imports.
    from betse.exceptions import BetseTestHookException
    from betse.util.test.pytest.pytabc import SerialTestABC

    # For each list of fixtures requested by this test...
    for fixture_defs in item._fixtureinfo.name2fixturedefs.values():
        # For each such fixture...
        for fixture_def in fixture_defs:
            # True only if this fixture is parametrized serially (i.e., is
            # decorated by the @serialize_parametrized_fixture decorator).
            is_fixture_parametrized_serially = getattr(
                fixture_def.func,
                '_betse_is_fixture_parametrized_serially',
                False)
            # print('funcinfo: {}'.format(fixture_def))
            # print('funcinfo.func: {}'.format(fixture_def.func))
            # print('funcinfo.func.is: {}'.format(is_fixture_parametrized_serially))

            # If this fixture is parametrized serially, propagate this setting
            # to this test by dynamically marking this test accordingly.
            if is_fixture_parametrized_serially:
                item.keywords['serialize_parametrized_test'] = True

    # If this test is parametrized serially, do so *AFTER* propagating this
    # setting from fixtures requested by this test above.
    if 'serialize_parametrized_test' in item.keywords:
        # If this test is *NOT* parametrized, raise an exception.
        if not hasattr(item, '_genid'):
            raise BetseTestHookException(
                'Unparametrized test "{}" marked as serial.'.format(item.name))

        # Test callable (i.e., underlying function or method object).
        test_callable = item.obj

        # Unique identifier of the first failing parameters passed to this test
        # if any or "None" otherwise.
        first_failing_param_id = getattr(
            test_callable, '_betse_first_failing_param_id', None)

        # If any parameters passed to this test failed, mark these subsequently
        # passed parameters as also xfailing.
        if first_failing_param_id is not None:
            pytest.xfail(
                'Prior serial test parametrization "{}" failed.'.format(
                    first_failing_param_id))
    #FIXME: This should *REALLY* simply be implemented as a new
    #@serialize_class decorator, much like @serialize_parametrized_test above.

    # If this is an unparametrized test intended to be run serially, do so.
    elif SerialTestABC.is_test_serial(item):
        # Object to which this test method is bound.
        test_instance = item.parent

        # Name of the first previously declared test method to fail in this
        # test method's class if any or "None" otherwise.
        first_failure_method_name = getattr(
            test_instance, '_first_failure_method_name', None)

        # If a previously declared test failed, mark this subsequently declared
        # test as also xfailing.
        if first_failure_method_name is not None:
            pytest.xfail('Prior serial test "{}" failed.'.format(
                test_instance._first_failure_method_name))


def pytest_runtest_makereport(item, call):
    '''
    Hook run immediately *after* the passed test returned the passed result.

    Specifically:

    * If this is a **serial test** (i.e., test method bound to an instance of
      the the :class:`betse.util.test.pytest.pytabc.SerialTestABC` superclass)
      that failed, this hook records this failure for subsequent analysis by
      the :func:`pytest_runtest_setup` hook.

    Parameters
    ----------
    item : pytest.main.Item
        Metadata encapsulating this test callable (e.g., function, method).
    call : pytest.runner.CallInfo
        Metadata encapsulating the value returned or exception raised by
        calling this test callable.

    See Also
    --------
    https://pytest.org/latest/example/simple.html#incremental-testing-test-steps
        Official py.test code snippet inspiring this implementation.
    '''

    # Defer heavyweight imports.
    from betse.util.test.pytest.pytabc import SerialTestABC

    # If this test failed...
    if call.excinfo is not None:
        # If this is a parametrized test intended to be run serially...
        if 'serialize_parametrized_test' in item.keywords:
            # Test callable (i.e., underlying function or method object).
            test_callable = item.obj

            # Record the unique identifiers of these test parameters for
            # subsequent use by the pytest_runtest_setup() hook.
            test_callable._betse_first_failing_param_id = item._genid
        #FIXME: This should *REALLY* simply be implemented as a new
        #@class_serial decorator, much like @parametrize_serial above.
        # If this is an unparametrized test intended to be run serially...
        elif SerialTestABC.is_test_serial(item):
            # Object to which this test method is bound.
            test_parent = item.parent

            # Record the name of this test for subsequent use by the
            # pytest_runtest_setup() hook.
            test_parent._first_failure_method_name = item.name



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all matplotlib-based layer subclasses  spatially
plotting onto the cell cluster.
'''

#FIXME: The current approach to implementing animation overlays is totally
#fundamentally flawed. We currently attempt to provide a crude form of plot
#composition (i.e., merging two or more types of plots together into a single
#plot) by adding new booleans to the "AnimCellsABC" base class (e.g.,
#"is_current_overlayable") -- a fundamentally unwieldy and ultimately
#unworkable approach. By definition, you cannot provide true composability from
#within a single class hierarchy. Instead, we need to split the specific
#process of generating different types of artists (e.g., mesh plots, stream
#plots) from the general process of animating and saving frames and plots as
#follows:
#
#* Refactor all concrete subclasses of "AnimCellsABC" into one or more
#  subclasses of "LayerCellsABC" instead, which may then be instantiated and
#  composed together into a new "layers" list passed to
#  CellsLayerABC.__init__(). For example:
#  * Split the existing "AnimGapJuncTimeSeries" subclass into:
#    * A new "CellsLayerGapJunc" subclass plotting *ONLY* the gap junction
#      open state as a "LineCollection" overlay. This layer subclass would
#      probably only be used for this specific purpose.
#    * A new "CellsLayerTimeSeries" subclass plotting *ONLY* an arbitrary
#      time series for the cell cluster as a mesh plot underlay. Clearly, this
#      layer subclass would be extensively reused elsewhere as well.
#* Replace all current overlay functionality in "AnimCellsABC" with "layers".
#* Refactor the configuration file from the current hard-coded non-composable
#  approach to a dynamic list-based approach permitting zero or more
#  user-defined animations, each consisting of one or more stock BETSE-defined
#  layers, to be defined. Users would then be able to construct arbitrarily
#  simple or complex animations as required.
#
#So, yes. It's quite a bit of work. But it's absolutely essential as well,
#particularly for implementing a general-purpose BETSE GUI.

# ....................{ IMPORTS                           }....................
from abc import ABCMeta, abstractmethod
from betse.exceptions import BetseSimVisualLayerException
from betse.util.io.log import logs
from betse.util.py import pyref
from betse.util.type import types
from betse.util.type.iterable import iterget
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.obj import objtest
from betse.util.type.types import (
    type_check, IterableTypes, SequenceOrNoneTypes,)
from matplotlib.cm import ScalarMappable

# ....................{ SUPERCLASS                        }....................
class LayerCellsABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all subclasses spatially plotting a single feature
    of the cell cluster for a parent visual.

    Each subclass of this class plots the spatial distribution of a single
    modelled variable (e.g., membrane voltage) for one or more simulation time
    steps. Each instance of the higher-level
    :class:`betse.science.visual.visabc.VisualCellsABC` abstract base class
    contains one or more instances of subclasses of this lower-level class.

    Separating low-level layer logic from high-level visual logic (e.g., frame
    iteration, video compression) enables composition between otherwise
    unrelated types. Thanks to layers, two or more types of visuals may be
    trivially composed into a unique third type of visual with *no*
    modification to existing layers or visuals.

    Attributes
    ----------
    _is_layered : bool
        ``True`` only if the :meth:`layer` method has been called at least once
        for this layer instance.
    _phase : SimPhase
        Current simulation phase *or* ``None`` if the :meth:`prep` method has
        yet to be called. Note that this attribute is also accessible via the
        :meth:`_visual.phase` property and is thus technically redundant. For
        both convenience and orthogonality with similar ``_phase`` attributes
        elsewhere in the codebase, this attribute is nonetheless defined and
        should be used in place of the :meth:`_visual.phase` property.
    _visual : VisualCellsABC
        Plot or animation to layer onto *or* ``None`` if the :meth:`prep`
        method has yet to be called.
    _zorder : int
        **Z-order** (i.e., positive integer ordering artist drawing, such that
        artists with larger z-orders are drawn above artists with smaller
        z-orders) of all artists plotted by this layer. While this zorder is
        *not* strictly enforced by this abstract base class, layer subclasses
        are encouraged to voluntarily pass the ``zorder=self._zorder`` option
        to all matplotlib axis-specific artist creation methods.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self) -> None:
        '''
        Initialize this layer.

        This method intentionally accepts *no* parameters except constants
        parametrizing this layer's behaviour. In particular, this method
        accepts references to neither the parent
        :class:`betse.science.visual.visabc.VisualCellsABC` instance containing
        this layer instance *nor* any other instances also contained by that
        parent instance (e.g., Matplotlib figure or axes objects). Why? Because
        plotters are instantiated by callers *before* their parent
        :class:`VisualCellsABC` instances are instantiated.

        See Also
        ----------
        :meth:`layer`
            Further details on class design.
        '''

        # Default instance attributes.
        self._is_layered = False
        self._phase = None
        self._visual = None
        self._zorder = None


    @type_check
    def prep(
        self,
        visual: 'betse.science.visual.visabc.VisualCellsABC',
        zorder: int,
    ) -> None:
        '''
        Prepare this layer to be layered onto the passed plot or animation.

        Parameters
        ----------
        visual : VisualCellsABC
            Plot or animation to layer onto.
        zorder : int
            **Z-order** (i.e., positive integer ordering artist drawing, such
            that artists with larger z-orders are drawn above artists with
            smaller z-orders) of all artists plotted by this layer.
        '''

        # Classify this visual with a weak rather than strong (the default)
        # reference, thereby avoiding circular references and the resulting
        # complications thereof (e.g., increased memory overhead).  Since the
        # parent plot or animation necessarily lives significantly longer than
        # this layer, no complications arise. Ergo, this attribute *ALWAYS*
        # yields this object (rather than non-deterministically yielding "None"
        # if this object is unexpectedly garbage-collected).
        self._visual = pyref.proxy_weak(visual)
        # self._visual = visual

        # Classify this zorder unmodified.
        self._zorder = zorder

        # Alias the current simulation phase to a convenience variable.
        self._phase = self._visual.phase

        # Ensure the next call to the layer() method calls the _layer_first()
        # rather than _layer_next() method, ensuring layers to be safely
        # reusable between multiple parent visuals.
        self._is_layered = False

    # ..................{ LAYERS                            }..................
    def layer(self) -> None:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        cell membrane voltage) for the current time step and each cell of the
        current cluster onto the figure axes of the current plot or animation.
        '''

        # If this method has yet to be called...
        if not self._is_layered:
            # Perform logic specific to this call.
            self._layer_first()

            # Prevent subsequent calls to this method from repeating this
            # logic.
            self._is_layered = True
        # Else, this method has been called at least once.
        else:
            # Perform logic specific to all subsequent calls.
            self._layer_next()


    # Layer subclasses are recommended but *NOT* required to reimplement this
    # empty method. Since simplistic layers plotting artists whose appearance
    # is constant across all time steps only require the abstract
    # _layer_first() method to be implemented, this method is concrete rather
    # than abstract.
    def _layer_next(self) -> None:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        cell membrane voltage) for the next simulation time step onto the
        figure axes of the current plot or animation.
        '''

        pass

    # ..................{ SUBCLASS                          }..................
    # Subclasses are required to implement the following abstract methods.

    @abstractmethod
    def _layer_first(self) -> None:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        cell membrane voltage) for the first simulation time step onto the
        figure axes of the current plot or animation.
        '''

        pass

# ....................{ SUBCLASSES ~ colorful             }....................
class LayerCellsColorfulABC(LayerCellsABC):
    '''
    Abstract base class of all subclasses spatially plotting a single modelled
    variable of the cell cluster (e.g., cell membrane voltage) whose values are
    mappable as colors onto the colorbar for a parent plot or animation.

    Attributes
    ----------
    _color_mappables : IterableTypes
        Iterable of all mappables internally cached and returned by the
        :meth:`color_mappables` property.
    _color_max : NumericSimpleTypes
        Maximum color value to be displayed by the parent visual's colorbar.
    _color_min : NumericSimpleTypes
        Minimum color value to be displayed by the parent visual's colorbar.
        Ignored if :attr:`_visual.conf.is_color_autoscaled` is ``True``.
        Minimum color value to be displayed by the colorbar. Ignored if
        :attr:`_conf.is_color_autoscaled` is ``True``.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self) -> None:
        '''
        Initialize this layer.
        '''

        # Initialize our superclass.
        super().__init__()

        # Default all instance attributes.
        self._color_mappables = None
        self._color_min = None
        self._color_max = None

    # ..................{ SUPERCLASS                        }..................
    #FIXME: Generalize this class to support layers recomputing color mappables
    #each time step. To do so:
    #
    #* Define a new abstract _layer_next_color_mappables() method.
    #* Refactor all subclasses to override that rather than the _layer_next()
    #  method. (Ugh. Apologies on that one.)
    #* Refactor all subclass implementations of _layer_next_color_mappables()
    #  to explicitly return None, for safety.
    #* Override the _layer_next() method here as follows:
    #
    #    def _layer_next(self) -> None:
    #
    #        # Iterable of mappables layered by the subclass for the current time
    #        # step if any or None otherwise.
    #        color_mappables = self._layer_next_color_mappables()
    #
    #        if color_mappables is not None:
    #            self._color_mappables = color_mappables
    #
    #Great. We're not quite done, however. The question then becomes: how do we
    #propagate this change of color mappables to the parent visual, which will
    #be required to reset its colorbar accordingly? *UGH.*

    def _layer_first(self) -> None:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        transmembrane voltage) for the first sampled time step onto the current
        visual's figure axes.

        This method internally calls the subclass
        :meth:`_layer_first_color_mappables` method and caches the
        :attr:`_color_mappables` iterable returned by that call.

        Caveats
        ----------
        **Subclasses should not override this method,** which preserves various
        invariants required for colorbar mapping. Instead, subclasses should
        *only* override the :meth:`_layer_first_color_mappables`,
        :meth:`_layer_next` method, and :meth:`_layer_next_color_mappables`
        methods as needed.
        '''

        # Color mappables layered by the subclass for the first time step.
        self._color_mappables = self._layer_first_color_mappables()

        # If no mappables were layered, raise an exception.
        if not self._color_mappables:
            raise BetseSimVisualLayerException(
                '_layer_first_color_mappables() returned no color mappables.')

        # Add a colorbar to this visual associated with these mappables.
        self._make_colorbar()


    def _layer_next(self) -> None:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        transmembrane voltage) for the next sampled time step onto the current
        visual's figure axes.

        This method internally calls the subclass
        :meth:`_layer_next_color_mappables` method and caches the
        :attr:`_color_mappables` iterable returned by that call.

        Caveats
        ----------
        **Subclasses should typically override this method,** which
        inefficiently recreates all color mappables each time step, which most
        layer subclasses do *not* require. To efficiently preserve the initial
        color mappables created and cached by the prior call to the
        :meth:`_layer_first` method, most layer subclasses should instead
        override the default implementation of this method rather than
        implementing the optional :meth:`_layer_next_color_mappables` method.
        '''

        # Color mappables layered by the subclass for the next time step.
        self._color_mappables = self._layer_next_color_mappables()

        # If no mappables were layered, raise an exception.
        if not self._color_mappables:
            raise BetseSimVisualLayerException(
                '_layer_next_color_mappables() returned no color mappables.')

        # Rescale these mappables to the previously established minimum and
        # maximum color values. Since matplotlib provides no convenient means
        # for efficiently replacing the existing colorbar added by the prior
        # call to the _layer_first() method *AND* since replacing this colorbar
        # would be generally undesirable, this colorbar is preserved as is.
        self._scale_color_mappables()

    # ..................{ SUBCLASS ~ mandatory              }..................
    # The following abstract methods *MUST* be implemented by subclasses.

    @abstractproperty
    def color_data(self) -> SequenceOrNoneTypes:
        '''
        Sequence of arbitrary dimensions of all possible color values for all
        time steps plotted by this layer *or* ``None`` if calculating these
        values is impractical (e.g., due to space or time constraints).

        If colorbar autoscaling is:

        * Disabled, this sequence is ignored.
        * Enabled *and* this sequence is:
          * ``None``, this sequence is ignored. In this case, the subclass is
            responsible for colorbar autoscaling.
          * Non-``None``, the colorbar is clipped to the minimum and maximum
            scalar values unravelled from this sequence.
        '''

        pass


    @abstractmethod
    def _layer_first_color_mappables(self) -> IterableTypes:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        transmembrane voltage) for the first sampled time step onto the current
        visual's figure axes, returning all matplotlib color mappables whose
        artists are to be mapped onto (i.e., coloured according to) this
        figure's colorbar.

        Returns
        ----------
        IterableTypes
            Iterable of all matplotlib color mappables, cached into the
            :attr:`_color_mappables` attribute by the parent
            :meth:`_layer_first` method. Note that, by matplotlib design, only
            the first mappable in this iterable is arbitrarily associated with
            this colorbar; all other mappables are ignored for this purpose.
        '''

        pass

    # ..................{ SUBCLASS ~ optional               }..................
    # The following non-abstract methods *MAY* be implemented by subclasses.

    def _layer_next_color_mappables(self) -> IterableTypes:
        '''
        Layer the spatial distribution of a single modelled variable (e.g.,
        transmembrane voltage) for the next sampled time step onto the current
        visual's figure axes, returning all matplotlib color mappables whose
        artists are to be mapped onto (i.e., coloured according to) this
        figure's colorbar.

        Design
        ----------
        All layer subclasses *must* implement either:

        * The :meth:`layer_next` method, in which case this layer is assumed to
          preserve the original color mappables returned by the initial call to
          the :meth:`_layer_first_color_mappables` method for each time step.
          This is the common (and most efficient) case.
        * This method, in which case this layer is assumed to behave as follows
          for each time step:

          * Destroy all color mappables created by the prior call to this or
            the :meth:`_layer_first_color_mappables` method.
          * Create and return new color mappables from the call to this method.

        Since the latter is substantially less efficient than the former,
        *only* layers recreating color mappables must implement this method;
        all other layers should implement the :meth:`layer_next` method.

        Returns
        ----------
        IterableTypes
            Iterable of all matplotlib color mappables, cached into the
            :attr:`_color_mappables` attribute by the parent
            :meth:`_layer_next` method.
        '''

        pass

    # ..................{ COLORS                            }..................
    def _make_colorbar(self) -> None:
        '''
        Add a colorbar to the parent visual's figure.

        This method automatically configures the range of colors displayed by
        this colorbar *and* associates all color mappables defined by this
        layer subclass with this colorbar.
        '''

        # Set the minimum and maximum colorbar values.
        self._set_color_range()

        # Scale all color mappables defined by this subclass by these values.
        self._scale_color_mappables()

        # First color mappable defined by this subclass.
        color_mappable_first = iterget.get_item_first(self._color_mappables)

        # Create a colorbar associated with this color mappable.
        self._visual.make_colorbar(color_mappable_first)


    def _set_color_range(self) -> None:
        '''
        Set the minimum and maximum color values to be displayed by the
        colorbar for the parent visual's figure.

        Specifically, these values are set to:

        * If colorbar autoscaling is enabled by this visual's configuration,
          the minimum and maximum values unravelled from the
          :meth:`color_data` Numpy array defined by this layer subclass.
        * Else, the minimum and maximum values hardcoded into this visual's
          configuration.
        '''

        # If colorbar autoscaling is enabled by this visual's configuration...
        if self._visual.conf.is_color_autoscaled:
            # One-dimensional Numpy array of all colors flattened from this
            # possibly multi-dimensional Numpy array of these values.
            color_data_flat = self.color_data.ravel()

            # Set the minimum and maximum colors to the minimum and maximum
            # values in this array.
            self._color_min = color_data_flat.min()
            self._color_max = color_data_flat.max()
            # self._color_min = np.ma.min(color_data_flat)
            # self._color_max = np.ma.max(color_data_flat)
        # Else, colorbar autoscaling is disabled. In this case, set the minimum
        # and maximum colors to those hardcoded into this configuration.
        else:
            self._color_min = self._visual.conf.color_min
            self._color_max = self._visual.conf.color_max

        # If these values are identical, coerce them to differ. Failing to do
        # so produces spurious visual artifacts in both the axes and colorbar.
        # if self._color_min == self._color_max:
        #     self._color_min = self._color_min - 1
        #     self._color_max = self._color_max + 1

        # Ensure sanity.
        assert types.is_numeric(self._color_min), (
            types.assert_not_numeric(self._color_min))
        assert types.is_numeric(self._color_max), (
            types.assert_not_numeric(self._color_max))


    #FIXME: DRY: this method exactly recapitulates the existing
    #VisualCellsABC._rescale_color_mappables() method. Excise one or the other.
    def _scale_color_mappables(self) -> None:
        '''
        Clip all color mappables defined by this layer subclass to the current
        minimum and maximum colormap values.

        This method has the beneficial side-effect of establishing the range of
        this visual's colorbar.
        '''

        # Log this attempt.
        logs.log_debug(
            'Rescaling "%s" colors to [%d, %d]...',
            self._visual.kind, self._color_min, self._color_max)

        # For each previously passed color mappable...
        for color_mappable in self._color_mappables:
            # If this is *NOT* actually a mappable, raise an exception.
            objtest.die_unless_instance(
                obj=color_mappable, cls=ScalarMappable)

            # Clip this mappable to the minimum and maximum colormap values.
            color_mappable.set_clim(self._color_min, self._color_max)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :func:`betse.util.py.pymodule.is_c` function.
'''

# ....................{ IMPORTS                           }....................

# ....................{ TESTS                             }....................
def test_is_c_unmonkeypatched() -> None:
    '''
    Unit test the :func:`betse.util.py.module.pymodule.is_c` function *without*
    monkeypatching the special PEP 302-specific ``__loader__`` attribute of
    tested modules.
    '''

    # Defer heavyweight imports.
    from betse.lib.numpy import numpys
    from betse.util.py.module import pymodule

    # Arbitrary Numpy submodule guaranteed to be implemented as a C extension.
    # Note that testing this particular C module is critical, as the core
    # codebase explicitly performs the same call at application startup.
    numpy_c_extension = numpys.get_c_extension()

    # Ensure that a C module is correctly detected as such.
    assert pymodule.is_c(numpy_c_extension) is True

    # Ensure that a pure-Python submodule is *NOT* detected to be a C
    # extension.
    assert pymodule.is_c(pymodule) is False


def test_is_c_monkeypatched(monkeypatch) -> None:
    '''
    Test all calls of the :func:`betse.util.py.module.pymodule.is_c` function
    by monkeypatching the special PEP 302-specific ``__loader__`` attribute of
    tested modules.

    This monkeypatch improves code coverage by exercising an infrequently used
    fallback codepath in this function.

    Parameters
    ----------
    monkeypatch : MonkeyPatch
        Builtin fixture object permitting object attributes to be safely
        modified for the duration of this unit test.
    '''

    # Imports deferred for safety.
    from betse.lib.numpy import numpys
    from betse.util.py.module import pymodule

    # Arbitrary Numpy submodule guaranteed to be implemented as a C extension.
    numpy_c_extension = numpys.get_c_extension()

    # Remove the PEP 302-specific "__loader__" attribute of each such module
    # before performing testing. Since this attribute need *NOT* exist, prevent
    # exceptions from being raised if this attribute does *NOT* exist.
    for module_object in (pymodule, numpy_c_extension):
        monkeypatch.delattr(module_object, name='__loader__', raising=False)

    # Defer to the existing base unit test.
    test_is_c_unmonkeypatched()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Dependency-specific REPL facilities.

This module defines functions specific to both official and optional
third-party REPL packages (e.g., :mod:`code`, :mod:`ptpython`).
'''

# ....................{ IMPORTS                           }....................
from betse.util.app.meta import appmetaone
from betse.cli.repl import replenv
from betse.lib import libs
from betse.util.io.log import logs
from betse.util.path import files
from betse.util.os.command import cmdexit
from betse.util.type.enums import make_enum
from betse.util.type.types import type_check

# ....................{ ENUMS                             }....................
# One-liners are happy liners.
ReplType = make_enum(
    class_name='ReplType',
    member_names=('first_available', 'ptpython', 'code',),
)
'''
Enumeration of all possible REPLs currently supported by this submodule.
'''

# ....................{ FUNCTIONS                         }....................
@type_check
def start_repl(repl_type: ReplType = ReplType.first_available) -> None:
    '''
    Start a REPL of the passed type.

    Parameters
    ----------
    repl_type : optional[ReplType]
        Type of REPL to prefer. If :data:`ReplType.first_available`, the set of
        all possible REPLs is iteratively searched for the first available
        REPL; else if this REPL is unavailable, the first available REPL is
        used. Defaults to :data:`ReplType.first_available`.
    '''

    if repl_type is ReplType.first_available:
        start_first_repl()
    elif repl_type is ReplType.ptpython:
        start_ptpython_repl()
    elif repl_type is ReplType.code:
        start_code_repl()
    else:
        logs.log_warning(
            'REPL type "{}" unrecognized. '
            'Deferring to first available REPL.'.format(repl_type.name))
        start_first_repl()


def start_first_repl() -> None:
    '''
    Start the first available REPL.
    '''

    if libs.is_runtime_optional('ptpython'):
        start_ptpython_repl()
    else:
        start_code_repl()


def start_ptpython_repl() -> None:
    '''
    Start a REPL based on the optional third-party :mod:`ptpython` package.

    If this package is unavailable, this function defers to the first available
    REPL with a non-fatal warning.
    '''

    # If "ptpython" is unavailable...
    if not libs.is_runtime_optional('ptpython'):
        # Log a non-fatal warning.
        logs.log_warning(
            '"ptpython" module not found. Deferring to first available REPL.')

        # Defer to the first available REPL.
        start_first_repl()

        # Get us out of here, Great Captain.
        return
    # Else, "ptpython" is available.

    # Log this invocation.
    logs.log_info('Starting "ptpython"-based REPL...')

    # Defer heavyweight imports.
    from ptpython.repl import embed

    # If the "ptpython" key is missing from the dictionary of history
    # filenames, then default to no history file. This prevents the readline
    # history files being corrupted by ptpython's unique format.
    history_filename = appmetaone.get_app_meta().get_repl_history_filename(
        'ptpython')

    # Run this REPL.
    try:
        embed(
            globals=None,
            locals=replenv.repl_env,
            history_filename=history_filename,
        )
    # When this REPL halts with error, reraise this exception.
    except SystemExit as exception:
        if cmdexit.is_failure(exception.exception):
            raise


def start_code_repl() -> None:
    '''
    Start a REPL based on the canonical :mod:`code` module.
    '''

    # Log this invocation, including the same helpful help line preceding the
    # official Python 3 REPL.
    logs.log_info('Starting "code"-based REPL...')
    logs.log_info(
        'Type "help", "copyright", "credits" or '
        '"license" for more information.')

    # Defer heavyweight imports.
    import code, readline

    # Absolute path of the file persisting a REPL-specific history of commands.
    # Note this REPL leverages a "readline"-style history file format.
    history_filename = appmetaone.get_app_meta().get_repl_history_filename(
        'readline')
    readline.set_history_length(1000)

    # If this file exists, deserialize this REPL's history from this file.
    if files.is_file(history_filename):
        logs.log_debug('Restoring REPL history from: %s', history_filename)
        readline.read_history_file(history_filename)

    # Run this REPL.
    try:
        code.interact(banner="", local=replenv.REPL_ENV)
    # When this REPL halts...
    except SystemExit as exception:
        # If this REPL halted with error, reraise this exception.
        if cmdexit.is_failure(exception.code):
            raise
        # Else, this REPL halted without error. Silently ignore this exception.
    # Serialize this REPL's history back to disk regardless of whether an
    # exception was raised.
    finally:
        logs.log_debug('Preserving REPL history to: %s', history_filename)
        readline.write_history_file(history_filename)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level integer facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseIntException
from betse.util.type import types
from betse.util.type.types import type_check

# ....................{ CONSTANTS                         }....................
BITS_PER_BYTE = 8
'''
Number of bits per byte.
'''

# ....................{ CONSTANTS ~ max                   }....................
BYTE_VALUE_MAX = 255
'''
Maximum value of unsigned bytes.
'''


INT_VALUE_MAX_32_BIT = 1 << 32
'''
Maximum value for integer variables of internal type ``Py_ssize_t`` on 32-bit
Python interpreters.

This value is suitable for comparison with :attr:`sys.maxsize`, the maximum
value of these variables on the current system.
'''

# ....................{ CONSTANTS ~ units ~ size          }....................
# Size denominations in base 2 rather than base 10, for mild efficiency.

KiB = 1 << 10
'''
Number of bytes in one kibibyte (i.e., ``2**10``).
'''


MiB = 1 << 20
'''
Number of bytes in one mebibyte (i.e., ``2**20``).
'''


GiB = 1 << 30
'''
Number of bytes in one gibibyte (i.e., ``2**30``).
'''


TiB = 1 << 40
'''
Number of bytes in one tebibyte (i.e., ``2**40``).
'''

# ....................{ EXCEPTIONS                        }....................
def die_unless(*objects) -> None:
    '''
    Raise an exception unless all passed objects are integers.

    Parameters
    ----------
    objects : tuple
        Tuple of all objects to be validated.

    Raises
    ----------
    BetseIntException
        If any passed object is _not_ an integer.
    '''

    for obj in objects:
        if not types.is_int(obj):
            raise BetseIntException(
                'Object "{}" not an integer.'.format(obj))

# ....................{ TESTERS ~ range                   }....................
@type_check
def die_unless_positive(*numbers: int, label: str = 'Integer') -> None:
    '''
    Raise an exception prefixed by the passed label unless all passed objects
    are positive integers.

    Parameters
    ----------
    numbers : tuple
        Tuple of all integers to be validated.
    label : optional[str]
        Human-readable label prefixing exception messages raised by this
        method. Defaults to a sensible string.

    Raises
    ----------
    BetseIntException
        If any passed integer is non-positive.
    '''

    # For each passed integer...
    for number in numbers:
        # If this integer is non-positive, raise an exception.
        if number <= 0:
            raise BetseIntException(
                '{} "{}" not positive.'.format(label.capitalize(), number))

# ....................{ TESTERS ~ size                    }....................
@type_check
def die_unless_byte(*numbers: int, label: str = 'Integer') -> None:
    '''
    Raise an exception prefixed by the passed label unless all passed objects
    are in the range ``[0, 255]``.

    Parameters
    ----------
    numbers : tuple
        Tuple of all integers to be validated.
    label : optional[str]
        Human-readable label prefixing exception messages raised by this
        method. Defaults to a sensible string.

    Raises
    ----------
    BetseIntException
        If any passed integer is *not* in the range ``[0, 255]``.
    '''

    # For each passed integer...
    for number in numbers:
        # If this integer is non-positive, raise an exception.
        if not is_byte(number):
            raise BetseIntException(
                '{} "{}" not a byte (i.e., not in the range [0, 255]).'.format(
                    label.capitalize(), number))

# ....................{ TESTERS ~ parity                  }....................
@type_check
def is_even(number: int) -> bool:
    '''
    ``True`` only if the passed integer is even.
    '''

    return number % 2 == 0


@type_check
def is_odd(number: int) -> bool:
    '''
    ``True`` only if the passed integer is odd.
    '''

    return number % 2 == 1

# ....................{ TESTERS ~ size                    }....................
@type_check
def is_byte(number: int) -> bool:
    '''
    ``True`` only if the passed integer is in the range ``[0, 255]``.

    Caveats
    ----------
    Equivalently, this function returns ``True`` only if this integer could
    technically occupy no more space than that required by a byte. Since
    integers in Python occupy a dynamic amount of space only loosely related to
    the current value they contain, a ``True`` return value does *not*
    necessarily imply this integer to consume exactly a byte's worth of space.
    '''

    # That was deceptively easy. Cue: "It's a trap!"
    return 0 <= number <= 255



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level facilities for **pipelining** (i.e., iteratively exporting)
post-simulation files containing raw simulation data, typically as plaintext
spreadsheets in comma-separated value (CSV) format.
'''

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.lib.numpy import nparray, npcsv
from betse.science.config.export.confexpcsv import SimConfExportCSV
from betse.science.math import mathunit
from betse.science.phase.phasecls import SimPhase
from betse.science.enum.enumphase import SimPhaseKind
from betse.science.phase.require import phasereqs
from betse.science.pipe.export.pipeexpabc import SimPipeExportABC
from betse.science.pipe.piperun import piperunner
from betse.science.visual.plot.plotutil import cell_ave
# from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from betse.util.type.descriptor.descs import classproperty_readonly
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from betse.util.type.types import (
    type_check, NumpyArrayType, SequenceTypes, StrOrNoneTypes,)

# ....................{ SUBCLASSES                        }....................
class SimPipeExportCSVs(SimPipeExportABC):
    '''
    **Post-simulation CSV export pipeline** (i.e., class iteratively creating
    all post-simulation plaintext spreadsheets in comma-separated value (CSV)
    format requested by the current simulation configuration).
    '''

    # ..................{ SUPERCLASS ~ properties           }..................
    @classproperty_readonly
    def _NOUN_SINGULAR(cls) -> str:
        return 'CSV file'

    # ..................{ SUPERCLASS                        }..................
    @type_check
    def iter_runners_conf(self, phase: SimPhase) -> SequenceTypes:
        return phase.p.csv.csvs_after_sim

    @type_check
    def _is_enabled(self, phase: SimPhase) -> bool:
        return phase.p.csv.is_after_sim_save

    # ..................{ EXPORTERS ~ cell                  }..................
    #FIXME: The requirements list should (arguably) be refined from the
    #coarse-grained "SOLVER_FULL" requirement to the exact list of fine-grained
    #requirements required by this exporter.
    @piperunner(
        categories=('Single Cell', 'Time Series'),
        requirements=phasereqs.SOLVER_FULL,
    )
    def export_cell_series(
        self, phase: SimPhase, conf: SimConfExportCSV) -> None:
        '''
        Save a plaintext file in comma-separated value (CSV) format containing
        several cell-specific time series (e.g., ion concentrations, membrane
        voltages, voltage-gated ion channel pump rates) for the single cell
        indexed ``plot cell index`` in the current simulation configuration.
        '''

        # 0-based index of the cell to serialize time data for.
        cell_index = phase.p.visual.single_cell_index

        # Sequence of key-value pairs containing all simulation data to be
        # exported for this cell, suitable for passing to the
        # OrderedArgsDict.__init__() method calleb below.
        csv_column_name_values = []

        # One-dimensional Numpy array of null data of the required length,
        # suitable for use as CSV column data for columns whose corresponding
        # simulation feature (e.g., deformations) is disabled.
        column_data_empty = np.zeros(len(phase.sim.time))

        # ................{ TIME STEPS                      }..................
        csv_column_name_values.extend(('time_s', phase.sim.time))

        # ................{ VMEM                            }..................
        csv_column_name_values.extend(
            ('Vmem_mV', self._get_cell_times_vmems(phase)))

        # ................{ VMEM ~ goldman                  }..................
        if phase.p.GHK_calc:
            vm_goldman = mathunit.upscale_units_milli([
                vm_GHK_time_cells[cell_index]
                for vm_GHK_time_cells in phase.sim.vm_GHK_time])
        else:
            vm_goldman = column_data_empty

        csv_column_name_values.extend(('Goldman_Vmem_mV', vm_goldman))

        # ................{ Na K PUMP RATE                  }..................
        if phase.p.is_ecm:
            pump_rate = [
                pump_array[phase.cells.cell_to_mems[cell_index][0]]
                for pump_array in phase.sim.rate_NaKATP_time]
        else:
            pump_rate = [
                pump_array[cell_index]
                for pump_array in phase.sim.rate_NaKATP_time]

        csv_column_name_values.extend((
            'NaK-ATPase_Rate_mol/m2s', pump_rate))

        # ................{ ION CONCENTRATIONS              }..................
        # Create the header starting with cell concentrations.
        for i in range(len(phase.sim.ionlabel)):
            csv_column_name = 'cell_{}_mmol/L'.format(
                phase.sim.ionlabel[i])
            cc_m = [arr[i][cell_index] for arr in phase.sim.cc_time]
            csv_column_name_values.extend((csv_column_name, cc_m))

        # ................{ MEMBRANE PERMEABILITIES         }..................
        # Create the header starting with membrane permeabilities.
        for i in range(len(phase.sim.ionlabel)):
            if phase.p.is_ecm:
                dd_m = [
                    arr[i][phase.cells.cell_to_mems[cell_index][0]]
                    for arr in phase.sim.dd_time
                ]
            else:
                dd_m = [arr[i][cell_index] for arr in phase.sim.dd_time]

            csv_column_name = 'Dm_{}_m2/s'.format(phase.sim.ionlabel[i])
            csv_column_name_values.extend((csv_column_name, dd_m))

        # ................{ TRANSMEMBRANE CURRENTS          }..................
        if phase.p.is_ecm:
            Imem = [
                memArray[phase.cells.cell_to_mems[cell_index][0]]
                for memArray in phase.sim.I_mem_time]
        else:
            Imem = [
                memArray[cell_index]
                for memArray in phase.sim.I_mem_time
            ]

        csv_column_name_values.extend(('I_A/m2', Imem))

        # ................{ HYDROSTATIC PRESSURE            }..................
        p_hydro = [arr[cell_index] for arr in phase.sim.P_cells_time]
        csv_column_name_values.extend(('HydroP_Pa', p_hydro))

        # ................{ OSMOTIC PRESSURE                }..................
        if phase.p.deform_osmo:
            p_osmo = [
                arr[cell_index] for arr in phase.sim.osmo_P_delta_time]
        else:
            p_osmo = column_data_empty

        csv_column_name_values.extend(('OsmoP_Pa', p_osmo))

        # ................{ DEFORMATION                     }..................
        if (
            phase.p.deformation and
            phase.kind is SimPhaseKind.SIM
        ):
            # Extract time-series deformation data for the plot cell:
            dx = nparray.from_iterable([
                arr[cell_index] for arr in phase.sim.dx_cell_time])
            dy = nparray.from_iterable([
                arr[cell_index] for arr in phase.sim.dy_cell_time])

            # Get the total magnitude.
            disp = mathunit.upscale_coordinates(np.sqrt(dx ** 2 + dy ** 2))
        else:
            disp = column_data_empty

        csv_column_name_values.extend(('Displacement_um', disp))

        # ................{ CSV EXPORT                      }..................
        # Ordered dictionary mapping from CSV column names to data arrays.
        csv_column_name_to_values = OrderedArgsDict(*csv_column_name_values)

        # Export this data to this CSV file.
        npcsv.write_csv(
            filename=self._get_csv_filename(
                phase=phase, basename_sans_filetype='ExportedData'),
            column_name_to_values=csv_column_name_to_values,
        )

    # ..................{ EXPORTERS ~ cell : vmem           }..................
    #FIXME: This exporter appears to currently be broken for the non-ECM
    #case, despite working for the ECM case. The exception being raised is:
    #
    #    BetseSequenceException: Column "FFT_Vmem" length 9 differs from length
    #    5 of prior columns.
    @piperunner(
        categories=('Single Cell', 'Voltage', 'FFT'),

        #FIXME: Eliminate this requirement after resolving the above issue.
        requirements=phasereqs.VOLTAGE_EXTRA,
    )
    def export_cell_vmem_fft(
        self, phase: SimPhase, conf: SimConfExportCSV) -> None:
        '''
        Save a plaintext file in comma-separated value (CSV) format containing
        the finite Fourier transform (FFT) of all transmembrane voltages for
        all sampled time steps spatially situated at the centre of the single
        cell indexed ``plot cell index`` in the current simulation
        configuration.
        '''

        # Number of sampled time steps.
        sample_size = len(phase.sim.time)

        # Time in seconds between each sampled time step.
        sample_spacing = phase.sim.time[1] - phase.sim.time[0]

        # One-dimensional Numpy array of all transmembrane voltages (Vmems) for
        # each sampled time step spatially situated at the centre of the single
        # cell indexed by the "plot cell index" entry for the passed phase.
        cell_times_vmems = self._get_cell_times_vmems(phase)

        cell_data = (1/sample_size) * (
            cell_times_vmems - np.mean(cell_times_vmems))

        # FFT of voltage.
        f_axis = np.fft.rfftfreq(sample_size, d=sample_spacing)
        fft_data_o = np.fft.rfft(cell_data)

        #FIXME: Numpy already provides a function for obtaining the magnitude of
        #a complex array: np.absolute(). The following math reduces to simply:
        #    fft_data = np.absolute(fft_data_o)
        fft_data = np.sqrt(np.real(fft_data_o)**2 + np.imag(fft_data_o)**2)
        # print('f_axis: {}'.format(f_axis))
        # print('fft_data: {}'.format(fft_data))

        # Ordered dictionary mapping from CSV column names to data arrays.
        csv_column_name_to_values = OrderedArgsDict(
            'frequency_Hz', f_axis,
            'FFT_Vmem', fft_data,
        )

        # Export this data to this CSV file.
        npcsv.write_csv(
            filename=self._get_csv_filename(
                phase=phase, basename_sans_filetype='ExportedData_FFT'),
            column_name_to_values=csv_column_name_to_values,
        )

    # ..................{ EXPORTERS ~ cells : vmem          }..................
    @piperunner(categories=('Cell Cluster', 'Voltage', 'Transmembrane',))
    def export_cells_vmem(
        self, phase: SimPhase, conf: SimConfExportCSV) -> None:
        '''
        Save one plaintext file in comma-separated value (CSV) format
        containing all transmembrane voltages (Vmem) upscaled and averaged from
        all cell membranes onto cell centres for each sampled time step of the
        current simulation phase.
        '''

        # Two-dimensional Numpy array of all transmembrane voltages.
        cells_times_vmems = mathunit.upscale_units_milli(
            phase.sim.vm_ave_time)

        # Export this data to this CSV file.
        self._export_cells_times_data(
            phase=phase,
            cells_times_data=cells_times_vmems,
            csv_column_name='Vmem [mV]',
            csv_dir_basename='Vmem2D_TextExport',
            csv_basename_prefix='Vmem2D_',
        )

    # ..................{ PRIVATE ~ getters                 }..................
    @type_check
    def _get_csv_filename(
        self,

        # Mandatory parameters.
        phase: SimPhase,
        basename_sans_filetype: str,

        # Optional parameters.
        dirname: StrOrNoneTypes = None,
    ) -> str:
        '''
        Absolute filename of a CSV file to be exported with the passed basename
        excluding suffixing ``.``-prefixed filetype (which this method appends
        to this basename as configured by the current simulation configuration)
        residing in the directory with the passed dirname.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        basename_sans_filetype : str
            Basename (excluding suffixing ``.``-prefixed filetype) of this
            file.
        dirname : StrOrNoneTypes
            Absolute pathname of the directory containing this file. If this
            directory does *not* already exist, this method creates this
            directory. Defaults to ``None``, in which case this pathname
            defaults to the top-level directory containing all CSV files
            exported for the current simulation phase.

        Returns
        ----------
        str
            Absolute filename of a CSV file to be exported with this basename.
        '''

        # If unpassed, default this directory to the top-level directory
        # containing all CSV files exported for this simulation phase.
        if dirname is None:
            dirname = phase.export_dirname

        # Create this directory if needed.
        dirs.make_unless_dir(dirname)

        # Basename of this file.
        basename = '{}.{}'.format(
            basename_sans_filetype, phase.p.csv.filetype)

        # Create and return the absolute filename of this file.
        return pathnames.join(dirname, basename)

    # ..................{ PRIVATE ~ exporters               }..................
    @type_check
    def _export_cells_times_data(
        self,
        phase: SimPhase,
        cells_times_data: SequenceTypes,
        csv_column_name: str,
        csv_dir_basename: str,
        csv_basename_prefix: str,
    ) -> None:
        '''
        Save one plaintext file in comma-separated value (CSV) format
        containing arbitrary simulation data spatially situated at cell centres
        for each sampled time step of the current simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        cells_times_data : ndarray
            Two-dimensional Numpy array of arbitrary simulation data spatially
            situated at the centres of all cells for all sampled time steps.
        csv_column_name : str
            Name of the column containing this data in all CSV-formatted files
            exported by this method.
        csv_dir_basename : str
            Basename of the directory containing all CSV-formatted files
            exported by this method.
        csv_basename_prefix : str
            Substring prefixing the basenames of all CSV-formatted files
            exported by this method.
        '''

        # Absolute pathname of the directory containing all CSV files
        # specifically exported by this method.
        csv_dirname = pathnames.join(
            phase.export_dirname, csv_dir_basename)

        # One-dimensional Numpy arrays of the X and Y coordinates
        # (respectively) of the centres of all cells.
        cell_centres_x = mathunit.upscale_coordinates(
            phase.cells.cell_centres[:,0])
        cell_centres_y = mathunit.upscale_coordinates(
            phase.cells.cell_centres[:,1])

        # For the 0-based index of each sampled time step...
        for time_step in range(len(phase.sim.time)):
            # Basename of the CSV-formatted file exported for this time step,
            # excluding suffixing "."-prefixed filetype.
            csv_basename_sans_filetype = '{}{}'.format(
                csv_basename_prefix, time_step)

            # Absolute filename of this CSV file.
            csv_filename = self._get_csv_filename(
                phase=phase,
                basename_sans_filetype=csv_basename_sans_filetype,
                dirname=csv_dirname,
            )

            # Ordered dictionary mapping from CSV column names to data arrays.
            csv_column_name_to_values = OrderedArgsDict(
                'x [um]', cell_centres_x,
                'y [um]', cell_centres_y,
                csv_column_name, cells_times_data[time_step],
            )

            # Export this data to this CSV file.
            npcsv.write_csv(
                filename=csv_filename,
                column_name_to_values=csv_column_name_to_values)

    # ..................{ PRIVATE ~ properties              }..................
    @type_check
    def _get_cell_times_vmems(self, phase: SimPhase) -> NumpyArrayType:
        '''
        One-dimensional Numpy array of all transmembrane voltages for each
        sampled time step spatially situated at the centre of the single cell
        indexed by the ``plot cell index`` entry specified by the passed
        simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        # 0-based index of the cell to serialize time data for.
        cell_index = phase.p.visual.single_cell_index

        if phase.p.is_ecm:
            cell_times_vmems = []
            for vm_at_mem in phase.sim.vm_time:
                vm_t = mathunit.upscale_units_milli(
                    cell_ave(phase.cells,vm_at_mem)[cell_index])
                cell_times_vmems.append(vm_t)
        else:
            cell_times_vmems = mathunit.upscale_units_milli(
                phase.sim.vm_time)

        return nparray.from_iterable(cell_times_vmems)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Voltage-gated potassium channel classes.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.channels.channelsabc import ChannelsABC
from betse.science.math import toolbox as tb
from betse.util.io.log import logs

# .................... BASE                               ....................
class VgKABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all Voltage-gated potassium channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel
    _hInf :
        Equillibrium value function for h-gates of channel
    _hTau :
        Time-constant function for h-gates of channel

    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize targeted voltage-gated sodium channels at the initial
        time step of the simulation based on the initial cell Vmems.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.
        '''

        if targets is None:

            self.targets = None

            if type(vm) == float or type(vm) == np.float64:
                self.data_length = 1

            else:
                self.data_length = len(vm)


        else:
            self.targets = targets
            self.data_length = len(self.targets)
            self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        self.v_corr = 0.0  # in experiments, the measurement junction voltage is about 10 mV

        if self.targets is None:

            V = vm * 1000 + self.v_corr

        else:
            V = vm[self.targets] * 1000 + self.v_corr

        self._init_state(V)

        self.ions = ['K']
        self.rel_perm = [1.0]

    def run(self, vm, p):
        '''
        Handle all targeted voltage-gated sodium channels by working with the passed
        user-specified parameters on the tissue simulation and cellular
        world for a time step.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.

        '''

        if self.targets is None:

            V = vm*1000 + self.v_corr

        else:

            V = vm[self.targets] * 1000 + self.v_corr


        self._calculate_state(V)

        self._implement_state(V, p)

    def _implement_state(self, V, p):

        # Update the 'm' and 'h' gating functions:
        self.update_mh(p, time_unit = self.time_unit)

        # calculate the open-probability of the channel:
        P = (self.m ** self._mpower) * (self.h ** self._hpower)

        if self.targets is None:

            self.P = P

        else:
            self.P = np.zeros(self.mdl)
            self.P[self.targets] = P


    @abstractmethod
    def _init_state(self, V):
        '''
        Initializes values of the m and h gates of the channel.
        '''
        pass

    @abstractmethod
    def _calculate_state(self, V):
        '''
        Calculates time-dependent values of the m and h gates of the channel.
        '''
        pass

# ....................{ SUBCLASS                           }....................

class Kv1p1(VgKABC):
    '''
    Kv1.1 model from Christie et al, cloned from rat brain, studied in xenopus.

    Kv1.1 are low-voltage activated (LVA) channels, expressed primarily in the central nervous system, and
    which open with small depolarizations at or below resting potential

    Reference: Christie MJ. et al. Expression of a cloned rat brain potassium channel in Xenopus oocytes.
    Science, 1989 Apr 14 , 244 (221-4).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1p1 ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -30.5000) / -11.3943))
        self.h = 1.0000 / (1 + np.exp((V - -30.0000) / 27.3943))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 2


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V - -30.5000) / -11.3943))
        self._mTau = 30.0000 / (1 + np.exp((V - -76.5600) / 26.1479))
        self._hInf = 1.0000 / (1 + np.exp((V - -30.0000) / 27.3943))
        self._hTau = 15000.0000 / (1 + np.exp((V - -160.5600) / -100.0000))

class Kv1p2(VgKABC):
    '''
    Kv1.2 model from  Sprunger et al.

    This channel produces well behaved action-potentials with a variety of vgNa channels. Good general-purpose
    vgK channel.

    Potassium voltage-gated channel Kv1.2 is a member of the shaker-related subfamily and belongs to the
    delayed rectifier class of channels, which allow nerve cells to efficiently re-polarize following an
    action potential. Kv1.2 is primarily found in the central nervous system.

    Reference: Sprunger LK. et al. Effects of charybdotoxin on K+ channel (KV1.2)
    deactivation and inactivation kinetics. Eur. J. Pharmacol., 1996 Oct 31 , 314 (357-64).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1.2')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000/(1+ np.exp(-(V +21.0000)/11.3943))
        self.h = 1.0000/(1+ np.exp((V + 22.0000)/11.3943))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000/(1+ np.exp(-(V +21.0000)/11.3943))
        self._mTau = 150.0000/(1+ np.exp((V + 67.5600)/34.1479))
        self._hInf = 1.0000/(1+ np.exp((V + 22.0000)/11.3943))
        self._hTau = 15000.0000/(1+ np.exp(-(V + 46.5600)/44.1479))

class Kv1p3(VgKABC):
    '''
    Kv1.3 model from Douglass et al 1990.

    Potassium voltage-gated channel Kv1.3 is a member of the shaker-related subfamily and belongs to the
    delayed rectifier class of channels, which allow nerve cells to efficiently re-polarize following an
    action potential. This channel is implicated in a host of non-neural activity, including autoimmune
    and inflammatory disorders, cell proliferation, and cancer development.

    reference:	Douglass J. et al. Characterization and functional expression of a rat genomic DNA
    clone encoding a lymphocyte potassium channel. J. Immunol., 1990 Jun 15 , 144 (4841-50).


    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1.3')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -14.1000) / -10.3000))
        self.h = 1.0000 / (1 + np.exp((V - -33.0000) / 3.7000))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V - -14.1000) / -10.3000))
        self._mTau = (-0.2840 * V) + 19.1600
        self._hInf = 1.0000 / (1 + np.exp((V - -33.0000) / 3.7000))
        self._hTau = (-13.7600 * V) + 1162.4000

class Kv1p4(VgKABC):
    '''
    Kv1.4 model from Stuhmer et al 1989.



    reference: W Stühmer et. al; EMBO J. 1989 Nov


    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1.4')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V + 21.7000) / -16.9000))
        self.h = 1.0000 / (1 + np.exp((V + 73.6000) / 12.8000))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V + 21.7000) / -16.9000))
        self._mTau = 3.0
        self._hInf = 1.0000 / (1 + np.exp((V + 73.6000) / 12.8000))
        self._hTau = 119.0

class Kv1p5(VgKABC):
    '''
    Kv1.5 model from

    This channel produces well behaved action-potentials with a variety of vgNa channels. Good general-purpose
    vgK channel.

    Kv1.5 underlies the cardiac ultra-rapid delayed rectifier potassium current (IKur) in humans.
    In addition, Kv1.5 channels are also expressed in many other organs, such as pulmonary arteries,
    brain, skeletal muscle, and have crucial function in cell cycle regulation.

    Reference: Philipson LH. et al. SequenceTypes and functional expression in Xenopus oocytes of a
    human insulinoma and islet potassium channel. Proc. Natl. Acad. Sci. U.S.A., 1991 Jan 1 , 88 (53-7).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1p5 ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -6.0000) / -6.4000))
        self.h = 1.0000 / (1 + np.exp((V - -25.3000) / 3.5000))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1.0000 / (1 + np.exp((V - -6.0000) / -6.4000))
        self._mTau = (-0.1163 * V) + 8.3300
        self._hInf = 1.0000 / (1 + np.exp((V - -25.3000) / 3.5000))
        self._hTau = (-15.5000 * V) + 1620.0000

class Kv1p6(VgKABC):
    '''
    Kv1.6 model from Grupe et al. 1990.



    Reference: A Grupe et. al; EMBO J. 1990 Jun

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: Kv1p6 ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(((V - (-20.800)) / (-8.100))))
        self.h =  1 / (1 + np.exp(((V - (-22.000)) / (11.390))))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1 / (1 + np.exp(((V - (-20.800)) / (-8.100))))
        self._mTau = 30.000 / (1 + np.exp(((V - (-46.560)) / (44.140))))
        self._hInf = 1 / (1 + np.exp(((V - (-22.000)) / (11.390))))
        self._hTau = 5000.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))

class Kv2p1(VgKABC):
    """
    Delayed rectifier potassium channel found widespread through many tissue types.

    Reference: 	VanDongen AM. et al. Alteration and restoration of K+ channel function by deletions
    at the N- and C-termini. Neuron, 1990 Oct , 5 (433-43).

    """


    def _init_state(self, V):
        """
        Run initialization calculation for m and h gates of the channel at starting Vmem value.
        """
        logs.log_info('You are using the vgK channel: Kv2p1 ')

        self.time_unit = 1.0e3

        self.vrev = -65  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(((V - (-9.200)) / (-6.600))))
        self.h = 1 / (1 + np.exp(((V - (-19.000)) / (5.000))))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1

    def _calculate_state(self, V):

        self._mInf = 1 / (1 + np.exp(((V - (-9.200)) / (-6.600))))
        self._mTau = 100.000 / (1 + np.exp(((V - (-46.560)) / (44.140))))
        self._hInf = 1 / (1 + np.exp(((V - (-19.000)) / (5.000))))
        self._hTau = 10000.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))

class Kv2p2(VgKABC):
    """
    Reference: F Schmalz et. al; Am. J. Physiol. 1998 May


    """


    def _init_state(self, V):
        """
        Run initialization calculation for m and h gates of the channel at starting Vmem value.
        """
        logs.log_info('You are using the vgK channel: Kv2p2 ')

        self.time_unit = 1.0e3

        self.vrev = -65  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(((V - (5.000)) / (-12.000))))
        self.h = 1 / (1 + np.exp(((V - (-16.300)) / (4.800))))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1

    def _calculate_state(self, V):

        self._mInf = 1 / (1 + np.exp(((V - (5.000)) / (-12.000))))
        self._mTau = 130.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))
        self._hInf = 1 / (1 + np.exp(((V - (-16.300)) / (4.800))))
        self._hTau = 10000.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))

class Kv3p1(VgKABC):
    """
 	J Rettig et. al; EMBO J. 1992 Jul

    """


    def _init_state(self, V):
        """
        Run initialization calculation for m and h gates of the channel at starting Vmem value.
        """
        logs.log_info('You are using the vgK channel: Kv3p1 ')

        self.time_unit = 1.0e3

        self.vrev = -65  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(((V - (18.700)) / (-9.700))))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

    def _calculate_state(self, V):

        self._mInf = 1 / (1 + np.exp(((V - (18.700)) / (-9.700))))
        self._mTau = 20.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))
        self._hInf = 1
        self._hTau = 1

class Kv3p2(VgKABC):
    """
    R Hernandez-Pineda et. al; J. Neurophysiol. 1999 Sep

    """


    def _init_state(self, V):
        """
        Run initialization calculation for m and h gates of the channel at starting Vmem value.
        """
        logs.log_info('You are using the vgK channel: Kv3p2 ')

        self.time_unit = 1.0e3

        self.vrev = -65  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp((V - -0.373267) / -8.568187))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 0

    def _calculate_state(self, V):

        self._mInf =  1 / (1 + np.exp((V - -0.373267) / -8.568187))
        self._mTau = 3.241643 + (19.106496 / (1 + np.exp((V - 19.220623) / 4.451533)))
        self._hInf = 1
        self._hTau = 1

class Kv3p3(VgKABC):


    """

     Kv3p3 is a rapidly inactivating, A-type inward/outward potassium current, implicated in rapid firing of
     neurons and is primarily found in Purkunjie neurons.

     Reference: Desai R. et al. Protein kinase C modulates inactivation of Kv3.3 channels.
     J. Biol. Chem., 2008 Aug 8 , 283 (22283-94).

    """

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the A-current K+ channel: Kv3p3')

        self.time_unit = 1.0e3

        # FIXME: IS THIS RIGHT?????
        self.vrev = 82.0  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp((V - 35) / -7.3))
        self.h =  0.25 + (0.75 / (1 + np.exp((V - (-28.293856)) / 29.385636)))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 2
        self._hpower = 1


    def _calculate_state(self, V):

        self._mInf = 1 / (1 + np.exp((V - 35) / -7.3))
        self._mTau = 0.676808 + (27.913114 / (1 + np.exp((V - 22.414149) / 9.704638)))
        self._hInf = 0.25 + (0.75 / (1 + np.exp((V - (-28.293856)) / 29.385636)))
        self._hTau = 199.786728 + (2776.119438 * np.exp(-V / 7.309565))

class Kv3p4(VgKABC):

    """

     Kv3p4 is a rapidly inactivating, A-type inward/outward potassium current.

     Reference: Vega-Saenz de Miera E. et al. Cloning of ShIII (Shaw-like) cDNAs encoding a novel
     high-voltage-activating, TEA-sensitive, type-A K+ channel. Proc. Biol. Sci., 1992 Apr 22 , 248 (9-18).

    """

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the K+ channel: Kv3p4')

        self.time_unit = 1.0e3

        self.vrev = -65.0  # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(((V - (-3.400)) / (-8.400))))
        self.h =  1 / (1 + np.exp(((V - (-53.320)) / (7.400))))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1

    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1 / (1 + np.exp(((V - (-3.400)) / (-8.400))))
        self._mTau = 10.000 / (1 + np.exp(((V - (4.440)) / (38.140))))
        self._hInf = 1 / (1 + np.exp(((V - (-53.320)) / (7.400))))
        self._hTau = 20000.000 / (1 + np.exp(((V - (-46.560)) / (-44.140))))

class K_Fast(VgKABC):
    '''
    "K Fast" model from Korngreen et al.

    Reference: Korngreen A. et al. Voltage-gated K+ channels in layer 5 neocortical pyramidal
    neurones from young rats: subtypes and gradients. J. Physiol. (Lond.), 2000 Jun 15 , 525 Pt 3 (621-39).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the vgK channel: K_Fast ')

        self.time_unit = 1.0e3

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the potassium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp(-(V + 47) / 29))
        self.h = 1 / (1 + np.exp(-(V + 56) / -10))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """
        self._mInf = 1 / (1 + np.exp(-(V + 47) / 29))
        self._mTau = (0.34 + 0.92 * np.exp(-((V + 71) / 59)**2))
        self._hInf = 1 / (1 + np.exp(-(V + 56) / -10))
        self._hTau = (8 + 49 * np.exp(-((V + 73) / 23)**2))

class KLeak(VgKABC):

    '''
    Simple potassium leak channel -- always open -- for substance modulation.

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using a K+ Leak channel')

        self.time_unit = 1.0

        self.vrev = -65     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = np.ones(self.data_length)
        self.h = np.ones(self.data_length)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 0
        self._hpower = 0


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1
        self._mTau = 1
        self._hInf = 1
        self._hTau = 1



        # print(self.modulator.mean())

class Kir2p1(VgKABC):
    '''
    Kir 2.1 model from Makary et al.

    This channel has a greater tendency to allow potassium to flow into a cell rather than out of a cell,
    probably participates in establishing action potential waveform and excitability of neuronal and muscle tissues.
    Mutations in this gene have been associated with Andersen syndrome, which is characterized by periodic paralysis,
    cardiac arrhythmias, and dysmorphic features

    Reference: Makary SM. et al. A difference in inward rectification and polyamine block and permeation between
     the Kir2.1 and Kir3.1/Kir3.4 K+ channels. J. Physiol. (Lond.), 2005 Nov 1 , 568 (749-66).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the inward rectifying K+ channel: Kir2p1')

        self.time_unit = 1.0e3

        self.vrev = -70.6     # reversal voltage used in model [mV]

        # initialize values of the m and h gates of the sodium channel based on m_inf and h_inf:
        self.m = 1 / (1 + np.exp((V - (-96.48)) / 23.26))
        self.h = 1 / (1 + np.exp((V - (-168.28)) / -44.13))

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 2

        # self.rectification = -1


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self._mInf = 1 / (1 + np.exp((V - (-96.48)) / 23.26))
        self._mTau = 3.7 + (-3.37 / (1 + np.exp((V - -32.9) / 27.93)))
        self._hInf = 1 / (1 + np.exp((V - (-168.28)) / -44.13))
        self._hTau = 0.85 + (306.3 / (1 + np.exp((V - -118.29) / -27.23)))















#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level facilities for **pipelining** (i.e., iteratively displaying and/or
exporting) post-simulation animations.
'''

#FIXME: This module would be a *GREAT* candidate for testing out Python 3.5-
#based asynchronicity and parallelization. Ideally, we'd be able to segregate
#the generation of each animation to its own Python process. Verdant shimmers!

# ....................{ IMPORTS                           }....................
from betse.science.config.export.visual.confexpvisanim import (
    SimConfExportAnimCells)
from betse.science.math.vector.veccls import VectorCellsCache
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require import phasereqs
from betse.science.pipe.export.pipeexpabc import SimPipeExportABC
from betse.science.pipe.piperun import piperunner
from betse.science.visual.anim.anim import (
    AnimGapJuncTimeSeries,
    # AnimMembraneTimeSeries,
    AnimVelocityIntracellular,
    AnimVelocityExtracellular,
    AnimFlatCellsTimeSeries,
)
from betse.science.visual.anim.animafter import AnimCellsAfterSolvingLayered
from betse.science.visual.layer.vector import lyrvecabc
from betse.science.visual.layer.vector.lyrvecdiscrete import (
    LayerCellsVectorDiscreteMembranesDeformed)
from betse.science.visual.layer.vector.lyrvecsmooth import (
    LayerCellsVectorSmoothGrids, LayerCellsVectorSmoothRegions)
from betse.science.visual.layer.vectorfield.lyrvecfldquiver import (
    LayerCellsFieldQuiverCells,
    LayerCellsFieldQuiverGrids,
    LayerCellsFieldQuiverMembranes,
)
from betse.science.visual.layer.vectorfield.lyrvecfldstream import (
    LayerCellsFieldStream)
from betse.util.type.descriptor.descs import classproperty_readonly
from betse.util.type.types import type_check, SequenceTypes

# ....................{ SUBCLASSES                        }....................
class SimPipeExportAnimCells(SimPipeExportABC):
    '''
    **Post-simulation animation pipeline** (i.e., class iteratively creating
    all post-simulation animations requested by the current simulation
    configuration).
    '''

    # ..................{ SUPERCLASS ~ properties           }..................
    @classproperty_readonly
    def _NOUN_SINGULAR(cls) -> str:
        return 'animation'

    # ..................{ SUPERCLASS ~ methods              }..................
    @type_check
    def iter_runners_conf(self, phase: SimPhase) -> SequenceTypes:
        return phase.p.anim.anims_after_sim

    @type_check
    def _is_enabled(self, phase: SimPhase) -> bool:
        return phase.p.anim.is_after_sim

    # ..................{ EXPORTERS ~ current               }..................
    @piperunner(
        categories=('Current Density', 'Intracellular',),
        requirements=phasereqs.ELECTRIC_CURRENT,
    )
    def export_currents_intra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all intracellular current densities for the cell cluster for
        all time steps.
        '''

        # Intracellular current density field.
        field = phase.cache.vector_field.currents_intra

        # Vector cache of all intracellular current density field magnitudes
        # over all time steps, spatially situated at cell centres.
        field_magnitudes = VectorCellsCache(
            phase=phase,
            times_cells_centre=field.times_cells_centre.magnitudes)

        # Layer sequence containing...
        layers = (
            # A lower layer animating these magnitudes.
            LayerCellsVectorSmoothRegions(vector=field_magnitudes),

            # A higher layer animating this field.
            LayerCellsFieldStream(field=field),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Intracellular Current',
            colorbar_title='Current Density [uA/cm2]',

            # Prefer an alternative colormap.
            colormap=phase.p.background_cm,
        )


    @piperunner(
        categories=('Current Density', 'Extracellular',),
        requirements=phasereqs.ELECTRIC_CURRENT_EXTRA,
    )
    def export_currents_extra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all extracellular current densities for the cell cluster
        environment over all sampled time steps.
        '''

        # Extracellular current density field over all sampled time steps.
        field = phase.cache.vector_field.currents_extra

        # Vector of all extracellular current density magnitudes over all time
        # steps, spatially situated at environmental grid space centres.
        field_magnitudes = VectorCellsCache(
            phase=phase,
            times_grids_centre=field.times_grids_centre.magnitudes)

        # Layer sequence containing...
        layers = (
            # A lower layer animating these magnitudes.
            LayerCellsVectorSmoothGrids(vector=field_magnitudes),

            # A higher layer animating this field.
            LayerCellsFieldStream(field=field),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Extracellular Current',
            colorbar_title='Current Density [uA/cm2]',

            # Prefer an alternative colormap.
            colormap=phase.p.background_cm,
        )

    # ..................{ EXPORTERS ~ deform                }..................
    @piperunner(
        categories=('Deformation', 'Total',),
        requirements=phasereqs.DEFORM,
    )
    def export_deform_total(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all **total cellular displacements** (i.e., summations of all
        cellular deformations due to galvanotropic and osmotic pressure body
        forces) for the cell cluster over all sampled time steps.
        '''

        # Total cellular displacements.
        field = phase.cache.vector_field.deform_total

        # Vector cache of all total cellular displacement magnitudes over all
        # time steps, spatially situated at cell centres.
        field_magnitudes = VectorCellsCache(
            phase=phase,
            times_cells_centre=field.times_cells_centre.magnitudes,
        )

        # Layer sequence containing...
        layers = (
            # A lower layer animating all cell displacement magnitudes.
            LayerCellsVectorDiscreteMembranesDeformed(vector=field_magnitudes),

            # A higher layer animating all cell displacement directionalities.
            LayerCellsFieldQuiverCells(field=field),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Deformation',
            colorbar_title='Displacement [um]',
        )

    # ..................{ EXPORTERS ~ electric              }..................
    @piperunner(
        categories=('Electric Field', 'Intracellular',),
        requirements=phasereqs.ELECTRIC_FIELD,
    )
    def export_electric_intra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all intracellular electric field lines for the cell cluster
        over all sampled time steps.
        '''

        # Intracellular electric field.
        field = phase.cache.vector_field.electric_intra

        # Vector cache of all intracellular electric field magnitudes over all
        # time steps, spatially situated at cell centres.
        field_magnitudes = VectorCellsCache(
            phase=phase,
            times_cells_centre=field.times_cells_centre.magnitudes)

        # Layer sequence containing...
        layers = (
            # A lower layer animating these magnitudes.
            LayerCellsVectorSmoothRegions(vector=field_magnitudes),

            # A higher layer animating this field.
            LayerCellsFieldQuiverCells(field=field),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Intracellular E Field',
            colorbar_title='Electric Field [V/m]',

            # Prefer an alternative colormap.
            colormap=phase.p.background_cm,
        )


    @piperunner(
        categories=('Electric Field', 'Extracellular',),
        requirements=phasereqs.ELECTRIC_FIELD_EXTRA,
    )
    def export_electric_extra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all extracellular electric field lines for the cell cluster
        environment over all sampled time steps.
        '''

        # Extracellular electric field over all sampled time steps.
        field = phase.cache.vector_field.electric_extra

        # Vector of all extracellular electric field magnitudes over all time
        # steps, spatially situated at environmental grid space centres.
        field_magnitudes = VectorCellsCache(
            phase=phase,
            times_grids_centre=field.times_grids_centre.magnitudes)

        # Layer sequence containing...
        layers = (
            # A lower layer animating these magnitudes.
            LayerCellsVectorSmoothGrids(vector=field_magnitudes),

            # A higher layer animating this field.
            LayerCellsFieldQuiverGrids(field=field),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Extracellular E Field',
            colorbar_title='Electric Field [V/m]',

            # Prefer an alternative colormap.
            colormap=phase.p.background_cm,
        )

    # ..................{ EXPORTERS ~ fluid                 }..................
    @piperunner(
        categories=('Fluid Flow', 'Intracellular',),
        requirements=phasereqs.FLUID,
    )
    def export_fluid_intra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all intracellular fluid flow field lines for the cell cluster
        over all sampled time steps.
        '''

        # Animate this animation.
        AnimVelocityIntracellular(
            phase=phase,
            conf=conf,
            kind='Velocity_gj',
            figure_title='Intracellular Fluid Velocity',
            colorbar_title='Fluid Velocity [nm/s]',
        )


    @piperunner(
        categories=('Fluid Flow', 'Extracellular',),
        requirements=phasereqs.FLUID_EXTRA,
    )
    def export_fluid_extra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all extracellular fluid flow field lines for the cell cluster
        environment over all sampled time steps.
        '''

        # Animate this animation.
        AnimVelocityExtracellular(
            phase=phase,
            conf=conf,
            kind='Velocity_ecm',
            figure_title='Extracellular Fluid Velocity',
            colorbar_title='Fluid Velocity [um/s]',
        )

    # ..................{ EXPORTERS ~ ion                   }..................
    @piperunner(
        categories=('Ion Concentration', 'Calcium',),
        requirements=phasereqs.ION_CALCIUM,
    )
    def export_ion_calcium_intra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all intracellular calcium ion (Ca2+) concentrations over all
        sampled time steps.
        '''

        # Layer sequence containing only a single layer animating these ion
        # concentrations.
        layers = (lyrvecabc.make_layer(
            phase=phase,
            vector=phase.cache.vector.ion_calcium_intra),)

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Cytosolic Ca2+',
            colorbar_title='Concentration [nmol/L]',
        )

    # ..................{ EXPORTERS ~ junction              }..................
    # This exporter is solver- and feature-agnostic.
    @piperunner(categories=('Gap Junction', 'Relative Permeability',))
    def export_gj_permeability(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all **gap junction connectivity states** (i.e., relative
        permeabilities of the gap junctions connecting all cell membranes) for
        the cell cluster over all sampled time steps.
        '''

        # Animate this animation.
        AnimGapJuncTimeSeries(
            phase=phase,
            conf=conf,
            time_series=phase.sim.gjopen_time,
            kind='Vmem_gj',
            figure_title='Gap Junction State over Vmem',
            colorbar_title='Voltage [mV]',
        )

    # ..................{ EXPORTERS ~ microtubules          }..................
    #FIXME: Restore after proper support for microtubules is implemented.
    # @piperunner(
    #     categories=('Microtubules', 'Coherence',),
    #     requirements=phasereqs.MICROTUBULE,
    # )
    # def export_microtubule(
    #     self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
    #     '''
    #     Animate the coherence of all cellular microtubules for the cell cluster
    #     over all sampled time steps.
    #     '''
    #
    #     # Sequence containing only a layer animating the vector field of all
    #     # cellular microtubules over all sampled time steps.
    #     layers = (LayerCellsFieldQuiverMembranes(
    #         field=phase.cache.vector_field.microtubule),)
    #
    #     # Animate these layers.
    #     AnimCellsAfterSolvingLayered(
    #         phase=phase,
    #         conf=conf,
    #         layers=layers,
    #         figure_title='Microtubule arrangement in cells',
    #     )

    # ..................{ EXPORTERS ~ pressure              }..................
    @piperunner(
        categories=('Pressure', 'Total',),
        requirements=phasereqs.PRESSURE_TOTAL,
    )
    def export_pressure_total(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all **cellular pressure totals** (i.e., summations of all
        cellular mechanical and osmotic pressures) for the cell cluster over
        all time steps.
        '''

        # Animate this animation.
        AnimFlatCellsTimeSeries(
            phase=phase,
            conf=conf,
            time_series=phase.sim.P_cells_time,
            kind='Pcell',
            figure_title='Pressure in Cells',
            colorbar_title='Pressure [Pa]',
        )


    @piperunner(
        categories=('Pressure', 'Osmotic',),
        requirements=phasereqs.PRESSURE_OSMOTIC,
    )
    def export_pressure_osmotic(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate the cellular osmotic pressure over all sampled time steps.
        '''

        # Animate this animation.
        AnimFlatCellsTimeSeries(
            phase=phase,
            conf=conf,
            time_series=phase.sim.osmo_P_delta_time,
            kind='Osmotic Pcell',
            figure_title='Osmotic Pressure in Cells',
            colorbar_title='Pressure [Pa]',
        )

    # ..................{ EXPORTERS ~ pump                   }..................
    # @piperunner(
    #     categories=('Ion Pump', 'Density Factor',),
    #     requirements={phasereqs.ELECTROOSMOSIS,},
    # )
    # def export_pump_density(self, conf: SimConfExportAnimCells) -> None:
    #     '''
    #     Animate all cell membrane ion pump density factors for the cell cluster
    #     over all sampled time steps.
    #     '''
    #
    #     # Animate this animation.
    #     AnimMembraneTimeSeries(
    #         phase=phase,
    #         conf=conf,
    #         time_series=phase.sim.rho_pump_time,
    #         kind='rhoPump',
    #         figure_title='Pump Density Factor',
    #         colorbar_title='mol fraction/m2',
    #     )

    # ..................{ EXPORTERS ~ voltage               }..................
    @piperunner(
        categories=('Voltage', 'Extracellular',),
        requirements=phasereqs.VOLTAGE_EXTRA,
    )
    def export_voltage_extra(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all extracellular voltages for the cell cluster environment
        over all sampled time steps.
        '''

        # Layer sequence containing only a single layer animating these
        # voltages.
        layers = (LayerCellsVectorSmoothGrids(
            vector=phase.cache.vector.voltage_extra),)

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Environmental Voltage',
            colorbar_title='Voltage [mV]',
        )


    @piperunner(categories=('Voltage', 'Transmembrane',))
    def export_voltage_membrane(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all transmembrane voltages (Vmem) for the cell cluster over all
        time steps.
        '''

        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=(phase.cache.vector.layer.voltage_membrane,),
            figure_title='Transmembrane Voltage',
            colorbar_title='Voltage [mV]',
        )


    @piperunner(
        categories=('Voltage', 'Polarity',),
        requirements=phasereqs.VOLTAGE_POLARITY,
    )
    def export_voltage_polarity(
        self, phase: SimPhase, conf: SimConfExportAnimCells) -> None:
        '''
        Animate all cellular voltage polarities for the cell cluster over all
        time steps.
        '''

        # Layer sequence containing...
        layers = (
            # A lower layer animating all transmembrane voltages.
            phase.cache.vector.layer.voltage_membrane,

            # A higher layer animating all transmembrane voltage polarities.
            LayerCellsFieldQuiverCells(
                field=phase.cache.vector_field.voltage_polarity),
        )

        # Animate these layers.
        AnimCellsAfterSolvingLayered(
            phase=phase,
            conf=conf,
            layers=layers,
            figure_title='Cell Vmem polarity',
            colorbar_title='Voltage [mV]',
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`pytest` functionality.

This submodule principally defines utility functions simplifying access to
:mod:`pytest` subpackages, submodules, and attributes unclassifiable under a
more specific submodule.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import ClassType, ModuleType

# ....................{ GETTERS                            }....................
def get_pytest_fixtures_submodule() -> ModuleType:
    '''
    Version-specific private :mod:`pytest` submodule providing fixture classes.

    Although private, this submodule publishes classes of public interest.
    Unfortunately, as this submodule *is* private, its fully-qualified name has
    changed across :mod:`pytest` versions. This function guaranteeably
    retrieves this submodule regardless of version.

    Specifically, under:

    * :mod:`pytest` >= 3.0.0, the private :mod:`_pytest.fixtures` submodule is
      imported and returned.
    * :mod:`pytest` < 3.0.0, the private :mod:`_pytest.python` submodule is
      imported and returned.
    '''

    # Attempt to import and return the newer "pytest" fixture module.
    try:
        from _pytest import fixtures as pytest_fixtures
    # Failing that, fallback to importing and returning the older such module.
    except ImportError:
        from _pytest import python as pytest_fixtures

    # In either case, return this module.
    return pytest_fixtures


def get_pytest_fixture_lookup_error_type() -> ClassType:
    '''
    Type of all :class:`FixtureLookupError` exceptions raised by :mod:`pytest`.

    Specifically, under:

    * :mod:`pytest` >= 3.0.0, the private
      :class:`_pytest.fixtures.FixtureLookupError` class is imported and
      returned.
    * :mod:`pytest` < 3.0.0, the private
      :class:`_pytest.python.FixtureLookupError` class is imported and
      returned.
    '''

    # Version-specific private `pytest` submodule providing fixture classes.
    pytest_fixtures_submodule = get_pytest_fixtures_submodule()

    # Return this exception class defined by this submodule.
    return pytest_fixtures_submodule.FixtureLookupError

# ....................{ OUTPUTTERS                         }....................
def output(*objs) -> None:
    '''
    Print all passed objects as is to standard output in a format mimicking
    that of standard :mod:`pytest` messages *without* logging these objects.

    This function is intended to be called *only* by :mod:`pytest`-specific
    fixtures, decorators, and helpers.

    This function is intentionally *not* named ``print()`` to avoid conflict
    with the builtin function of the same name.

    Examples
    --------
    .. code-block:: pycon

       >>> from betse.util.test.pytest import pytests
       >>> pytests.output('Ego, ergo simulare.')
       [py.test] Ego, ergo simulare.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command import cmds

    # Print these messages in a "py.test"-friendly format.
    print('[{}] {}'.format(cmds.get_current_basename(), ''.join(objs)))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Utility functions simplifying creation of **simulation phase requirement sets**
(i.e., immutable sets of simulation phase requirements, requiring zero or more
arbitrary requirements to be satisfied).
'''

# ....................{ IMPORTS                            }....................
from betse.science.phase.require.abc.phasereqabc import (
    SimPhaseRequirementABC,
    SimPhaseRequirement,
    SimPhaseRequirementEmbodied,
    SimPhaseRequirementBoolExpr,
    SimPhaseRequirementEnumExpr,
)
from betse.science.phase.require.abc.phasereqset import SimPhaseRequirements
# from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ MAKERS                             }....................
@type_check
def make_requirements(
    requirement: SimPhaseRequirementABC) -> SimPhaseRequirements:
    '''
    Immutable set of only one arbitrary simulation phase requirement.

    Parameters
    ----------
    requirement : SimPhaseRequirementABC
        Arbitrary requirement defining this set.
    '''

    # Create and return this requirement set from this requirement.
    return SimPhaseRequirements(iterable=(requirement,))


@type_check
def make_requirements_funcs(*args, **kwargs) -> SimPhaseRequirements:
    '''
    Immutable set of only one callable-based simulation phase requirement.

    All passed parameters are passed to the
    :meth:`SimPhaseRequirement.__init__` method as is.
    '''

    # Callable-based simulation phase requirement.
    requirement = SimPhaseRequirement(*args, **kwargs)

    # Create and return this requirement set from this requirement.
    return make_requirements(requirement)


@type_check
def make_requirements_bool_expr(*args, **kwargs) -> SimPhaseRequirements:
    '''
    Immutable set of only one string-based boolean simulation phase requirement.

    All passed parameters are passed to the
    :meth:`SimPhaseRequirementBoolExpr.__init__` method as is.
    '''

    # String-based boolean simulation phase requirement.
    requirement = SimPhaseRequirementBoolExpr(*args, **kwargs)

    # Create and return this requirement set from this requirement.
    return make_requirements(requirement)


@type_check
def make_requirements_enum_expr(*args, **kwargs) -> SimPhaseRequirements:
    '''
    Immutable set of only one string-based enumeration member simulation phase
    requirement.

    All passed parameters are passed to the
    :meth:`SimPhaseRequirementEnumExpr.__init__` method as is.
    '''

    # String-based enumeration member simulation phase requirement.
    requirement = SimPhaseRequirementEnumExpr(*args, **kwargs)

    # Create and return this requirement set from this requirement.
    return make_requirements(requirement)


@type_check
def make_requirements_ion(name: str, ion_name: str) -> SimPhaseRequirements:
    '''
    Immutable set of ionic simulation phase requirements, requiring both the
    full BETSE solver *and* a single ion type to be satisfied.

    Since simulation of ion concentrations requires the full BETSE solver, this
    requirement is actually the conjunction of that requirement *and* a custom
    requirement specific to this ion.

    Parameters
    ----------
    name : str
        Human-readable name of this requirement (e.g., "extracellular spaces").
        For usability, the first character of this name should typically be
        lower- rather than uppercase.
    ion_name : str
        Name of the required ion, equivalent to a key of the
        :attr:`Parameters.ions_dict` dictionary (e.g., ``Ca``, requiring
        calcium ions).
    '''

    # Avoid circular import dependencies.
    from betse.science.phase.require.phasereqs import SOLVER_FULL

    # Requirement that the passed ion is satisfied by a given phase.
    ion_requirement = SimPhaseRequirementEmbodied(
        name=name,
        is_satisfied_body=(
            'return phase.p.ions_dict[{!r}] == 1'.format(ion_name)),
        set_satisfied_body=(
            'phase.p.ions_dict[{!r}] = 1'.format(ion_name)),
    )

    # Create and return this requirement set from this requirement.
    return SOLVER_FULL | make_requirements(requirement=ion_requirement)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixture classes running BETSE CLI subcommands in the active Python interpreter.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import type_check, SequenceTypes
from betse_test.a90_func._fixture.clier import CLITester
from betse_test._fixture.simconf.simconfclser import SimConfTestABC

# ....................{ CLASSES                            }....................
class CLISimTester(object):
    '''
    BETSE CLI simulation test runner, exercising multiple subcommands of the
    BETSE CLI (i.e., ``betse``) in the active Python interpreter with the same
    temporary simulation configuration.

    Complex functional fixtures (e.g., :func:`betse_cli_sim`) typically return
    instances of this class to other fixtures and tests exercising multiple
    facets of the BETSE CLI.

    Attributes
    ----------
    cli_tester : CLITester
        BETSE CLI test runner, testing a single subcommand of the official
        BETSE CLI (i.e., ``betse``) in the active Python interpreter.
    sim_state: SimConfTestABC
        Test-specific object encapsulating a temporary simulation
        configuration file specific to the current test.

    See Also
    ----------
    :class:`CLITester`
        Further details on BETSE CLI execution.
    '''

    # ..................{ CONSTANTS                          }..................
    # The following tuples of argument tuples are intentionally declared as
    # class rather than global constants, dramatically simplifying usage by
    # fixtures and tests already passed an instance of this class.

    SUBCOMMANDS_SIM = (
        ('seed',),
        ('init',),
        ('sim',),
    )
    '''
    Tuple of all **simulation subcommand argument tuples** (i.e., tuple of one
    or more shell words which, when passed as command-line arguments to the
    ``betse`` command, perform a simulation-specific subcommand), excluding
    plotting.
    '''


    SUBCOMMANDS_PLOT = (
        ('plot', 'init'),
        ('plot', 'sim'),
    )
    '''
    Tuple of all **plotting subcommand argument tuples** (i.e., tuple of one or
    more shell words which, when passed as command-line arguments to the
    ``betse`` command, perform a plotting-specific subcommand).
    '''


    SUBCOMMANDS_TRY = SUBCOMMANDS_SIM + SUBCOMMANDS_PLOT
    '''
    Tuple of all **``try`` subcommand argument tuples** (i.e., tuple of one or
    more shell words which, when passed as command-line arguments to the
    ``betse`` command, perform the equivalent of the ``try`` subcommand).
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self, cli_tester: CLITester, sim_state: SimConfTestABC) -> None:
        '''
        Initialize this test runner.

        Parameters
        ----------
        cli_tester: CLITester
            BETSE CLI test runner, testing a single subcommand of the official
            BETSE CLI (i.e., ``betse``) in the active Python interpreter.
        sim_state: SimConfTestABC
            Test-specific object encapsulating a temporary simulation
            configuration file specific to the current test.
        '''

        # Classify these parameters.
        self.cli_tester = cli_tester
        self.sim_state = sim_state

    # ..................{ RUNNERS                            }..................
    @type_check
    def run_subcommands(
        self, *subcommands_args: SequenceTypes, is_overwrite_conf: bool = True
    ) -> None:
        '''
        Perform all BETSE CLI subcommands signified by the passed argument
        lists in the active Python process (in the passed order).

        To guarantee that each such subcommand efficiently reuses the same
        underlying simulation, this method implicitly:

        * Appends each such argument list with the absolute path of the
          simulation configuration file with which this test runner was
          initialized.
        * Temporarily changes the current working directory (CWD) to the
          directory containing this file.

        Caveats
        ----------
        **Order is significant.** Subcommands producing output required by
        subsequent subcommands as input should be passed first.

        Parameters
        ----------
        subcommands_args : tuple
            Tuple of sequences of **subcommand arguments** (i.e., one or more
            shell words comprising the BETSE CLI subcommand to be tested).
        is_overwrite_conf : optional[bool]
            If ``True``, all in-memory changes to the current simulation
            configuration are persisted back to disk. Defaults to ``True``.
        '''

        # If persisting all in-memory configuration changes back to disk, do so
        # *BEFORE* running the passed.subcommands requiring these changes.
        if is_overwrite_conf:
            self.overwrite_conf()

        # For each such subcommand...
        for subcommand_args in subcommands_args:
            # Run this subcommand, preventing the configuration file from being
            # re-overwritten. While technically safe, doing so would impose
            # unnecessary I/O inefficiencies.
            self.run_subcommand(
                *subcommand_args, is_overwrite_conf=False)


    @type_check
    def run_subcommand(
        self, *subcommand_args: str, is_overwrite_conf: bool = True
    ) -> None:
        '''
        Run the simulation-specific BETSE CLI subcommand signified by the
        passed argument list in the active Python process.

        This method implicitly:

        * Temporarily changes the current working directory (CWD) to the
          directory containing the desired simulation configuration file.
        * Appends this argument list by the basename of this file.

        The basename rather than absolute path of this file is passed to ensure
        that a fatal error is raised if the test requesting this fixture object
        failed to change the current working directory (CWD) to the directory
        containing this file. Why? Because operating outside of this directory
        encourages accidental permanent modification of the filesystem by tests
        and hence *must* be discouraged.

        Parameters
        ----------
        subcommand_args : tuple
            Tuple of **subcommand arguments** (i.e., one or more shell words
            comprising the BETSE CLI subcommand to be tested).
        is_overwrite_conf : optional[bool]
            If ``True``, all in-memory changes to the current simulation
            configuration are persisted back to disk. Defaults to ``True``.
        '''

        # Defer heavyweight imports.
        from betse.util.io.log import logs
        from betse.util.type.text.string import strjoin

        # If persisting all in-memory configuration changes back to disk, do so
        # *BEFORE* running the passed subcommands requiring these changes.
        if is_overwrite_conf:
            self.overwrite_conf()

        # Human-readable title to be embedded in a single-line terminal banner.
        subcommand_banner_title = strjoin.join_on(
            subcommand_args, delimiter=' ')

        # Log a single-line terminal banner embedding this title.
        logs.log_banner(title=subcommand_banner_title, padding='=')

        # Temporarily change the CWD to this simulation file's directory.
        with self.sim_state.context():
            # Append the absolute path of this runner's configuration file to
            # the passed tuple of arguments. While inefficient, converting this
            # tuple into a list would be even more inefficient.
            #
            # Avoid shell-quoting this path. Doing so unnecessarily adds an
            # additional level of quoting... which is bad.
            subcommand_args += (self.sim_state.conf_filename,)

            # Run this subcommand.
            self.cli_tester.run(*subcommand_args)

    # ..................{ RUNNERS ~ predefined               }..................
    def run_subcommands_sim(self) -> None:
        '''
        Perform all simulation-specific BETSE CLI subcommands excluding those
        pertaining to plotting.

        See Also
        ----------
        :attr:`SUBCOMMANDS_SIM`
            Tuple of all argument tuples comprising these subcommands.
        '''

        self.run_subcommands(*self.SUBCOMMANDS_SIM)


    def run_subcommands_try(self) -> None:
        '''
        Perform the equivalent of the BETSE CLI ``try`` subcommand in the
        active Python process.

        Specifically, this method performs all BETSE CLI subcommands performed
        by the ``try`` subcommand (in the same order).

        See Also
        ----------
        :attr:`SUBCOMMANDS_TRY`
            Tuple of all argument tuples comprising these subcommands.
        '''

        self.run_subcommands(*self.SUBCOMMANDS_TRY)

    # ..................{ OVERWRITERS                        }..................
    def overwrite_conf(self) -> None:
        '''
        Persist all in-memory changes to the current simulation configuration
        wrapper back to disk.
        '''

        self.sim_state.p.save_inplace()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string sequence** (i.e., non-string object implementing the
abstract base class `collections.abc.Sequence`) facilities.

See Also
----------
:func:`betse.util.type.types.is_sequence`
    Further details on what constitutes sequences and non-string sequences.
'''

#FIXME: Most if not all of the existing functionality provided by this
#submodule should be generalized to support higher-level iterable rather than
#lower-level sequence types and then shifted into the existing "iterables"
#submodule.

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseSequenceException
from betse.util.type import types
from betse.util.type.types import (
    type_check,
    CallableTypes,
    ContainerTypes,
    IterableTypes,
    MappingType,
    SequenceTypes,
    StrOrNoneTypes,
)

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_index(sequence: SequenceTypes, index: int) -> None:
    '''
    Raise an exception unless the passed index **indexes** (i.e., is a
    non-negative integer strictly greater than -1 and less than the length of)
    the passed sequence.

    Equivalently, this function raises an exception this index is either:

    * Less than 0.
    * Greater than or equal to the length of this sequence.

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to be validated.
    index : int
        0-based index to validate against this sequence.

    Raises
    ----------
    BetseSequenceException
        Unless this index indexes this sequence.
    '''

    # If this index does *NOT* index this sequence, raise an exception.
    if not is_index(sequence=sequence, index=index):
        raise BetseSequenceException(
            'Sequence index {} invalid (i.e., not in range [0, {}]).'.format(
                index, len(sequence)))

# ....................{ EXCEPTIONS ~ length               }....................
@type_check
def die_if_empty(
    sequence: SequenceTypes, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) if the passed sequence is **empty** (i.e., contains *no* items).

    Parameters
    ----------
    sequence: SequenceTypes
        Sequence to be validated.
    exception_message : StrOrNoneTypes
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseSequenceException
        If the sequence is empty.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objects

    # If this sequence is empty, raise an exception.
    if is_empty(sequence):
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'Sequence "{}" empty.'.format(
                objects.get_class_name_unqualified(sequence))

        # Raise this exception.
        raise BetseSequenceException(exception_message)



@type_check
def die_if_length_less_than(
    # Mandatory parameters.
    sequence: SequenceTypes,
    length: int,

    # Optional parameters.
    exception_message: StrOrNoneTypes = None,
) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) unless the length of the passed sequence is greater than or equal
    to the passed length (i.e., unless this sequence contains at least as many
    items as this length).

    Equivalently, this function raises an exception if this sequence contains
    strictly less than this many items.

    Parameters
    ----------
    sequence: SequenceType
        Sequence to be validated.
    length: int
        Minimum sequence length to test for.
    exception_message : StrOrNoneTypes
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseSequenceException
        If this sequence contains strictly less than this many items.
    '''

    # If this sequence contains less than this many items, raise an exception.
    if len(sequence) < length:
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'Sequence length {} < {}.'.format(
                len(sequence), length)

        # Raise this exception.
        raise BetseSequenceException(exception_message)


@type_check
def die_unless_length(
    # Mandatory parameters.
    sequence: SequenceTypes,
    length: int,

    # Optional parameters.
    exception_message: StrOrNoneTypes = None,
) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) unless the passed sequence is of the passed length.

    Equivalently, this function raises an exception if this sequence is *not*
    of this length.

    Parameters
    ----------
    sequence: SequenceType
        Sequence to be validated.
    length: int
        Exact sequence length to test for.
    exception_message : StrOrNoneTypes
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseSequenceException
        If this sequence is *not* of this length.
    '''

    # If this sequence is *NOT* of this length, raise an exception.
    if len(sequence) != length:
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'Sequence length {} not {}.'.format(
                len(sequence), length)

        # Raise this exception.
        raise BetseSequenceException(exception_message)

# ....................{ TESTERS                           }....................
@type_check
def is_empty(*sequences: SequenceTypes) -> bool:
    '''
    ``True`` only if all passed sequences are **empty** (i.e., contain no
    elements).

    Parameters
    ----------
    sequences : tuple[SequenceTypes]
        Tuple of all sequences to be tested.

    Returns
    ----------
    bool
        ``True`` only if these sequences are all empty.
    '''

    # all(). It is all awesome.
    #
    # To transparently support Numpy arrays, the length of this sequence *MUST*
    # be explicitly tested via the len() builtin rather than implicitly tested
    # as a boolean. On the latter, Numpy raises the following exception:
    #
    #     ValueError: The truth value of an array with more than one element is
    #     ambiguous. Use a.any() or a.all()
    return all(len(sequence) == 0 for sequence in sequences)


@type_check
def is_index(sequence: SequenceTypes, index: int) -> bool:
    '''
    ``True`` only if the passed index **indexes** (i.e., is a non-negative
    integer strictly greater than -1 and less than the length of) the passed
    sequence.

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to be tested.
    index : int
        0-based index to test against this sequence.

    Returns
    ----------
    bool
        ``True`` only if this index indexes this sequence.
    '''

    return 0 <= index < len(sequence)

# ....................{ TESTERS ~ type                    }....................
def is_sequence(*objs: object) -> bool:
    '''
    ``True`` only if all passed objects are **sequences** (i.e., of types
    conforming to but *not* necessarily subclassing the canonical
    :class:`collections.abc.Sequence` API).

    Parameters
    ----------
    objs : tuple[object]
        Tuple of all objects to be tested.

    Returns
    ----------
    bool
        ``True`` only if these objects are all sequences.

    See Also
    ----------
    :class:`collections.abc.Sequence`
        Further details.
    '''

    # all(). It is awesome.
    return all(isinstance(obj, SequenceTypes) for obj in objs)


#FIXME: Excise everywhere in favour of the more appropriate
#betse.lib.numpy.nparray.is_array() function -- which, admittedly, may need to
#be generalized to accept variadic parameters.
def is_numpy_array(*objs: object) -> bool:
    '''
    ``True`` only if all passed objects are **Numpy arrays** (i.e., instances
    of the :class:`numpy.ndarray` superclass).

    Parameters
    ----------
    objs : tuple[object]
        Tuple of all objects to be tested.

    Returns
    ----------
    bool
        ``True`` only if these objects are all Numpy arrays.
    '''

    # Avoid importing third-party packages at the top level, for safety.
    from numpy import ndarray

    # all(). It exceeds at all the tests.
    return all(isinstance(obj, ndarray) for obj in objs)

# ....................{ GETTERS                           }....................
@type_check
def get_index(sequence: SequenceTypes, index: int) -> object:
    '''
    Item with the passed index in this sequence if this index **indexes**
    (i.e., is a non-negative integer strictly greater than -1 and less than the
    length of) this sequence *or*  raise an exception otherwise (i.e., if this
    index does *not* index this sequence).

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to be inspected.
    index : int
        0-based index to return the item in this sequence of.

    Returns
    ----------
    object
        Item with this index in this sequence.

    Raises
    ----------
    BetseSequenceException
        If this index does *not* index this sequence.
    '''

    # If this index does *NOT* index this sequence, raise an exception.
    die_unless_index(sequence=sequence, index=index)
    # Else, this index indexes this sequence.

    # Return the item with this index in this sequence.
    return sequence[index]

# ....................{ GETTERS ~ items                   }....................
#FIXME: Generalize into a function of the same name accepting an iterable rather
#than sequence and shift into the "iterables" submodule. When doing so, rename
#the "item_satisfier" parameter to "predicate" for orthogonality and clarity.
@type_check
def get_items_satisfying(
    sequence: SequenceTypes, item_satisfier: CallableTypes) -> SequenceTypes:
    '''
    New non-string sequence containing only the proper subset of elements from
    the passed non-string sequence satisfying the passed callable.

    This method effectively performs a general-purpose global search for
    sequence elements.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does _not_ modify this sequence.
    item_satisfier : CallableTypes
        CallableTypes (e.g., function, lambda) accepting a single element of
        this sequence and returning only:

        * ``True`` if this element satisfies the desired requirements.
        * ``False`` otherwise.

    Returns
    ----------
    collections.Sequence
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Type of both the passed sequence and the sequence to be returned.
    sequence_type = type(sequence)

    # Return a generator-based shallow copy of this sequence.
    return sequence_type(item for item in sequence if item_satisfier(item))


#FIXME: Generalize into a function of the same name accepting an iterable rather
#than sequence and shift into the "iterables" submodule. When doing so, rename
#the "item_prefix" parameter to simply "prefix".
@type_check
def get_items_prefixed_by(
    sequence: SequenceTypes, item_prefix: str) -> SequenceTypes:
    '''
    New non-string sequence containing only the proper subset of elements from
    the passed non-string sequence that are strings prefixed by the passed
    string prefix.

    This method effectively performs a prefix-based global search for sequence
    elements.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does *not* modify this sequence.
    item_prefix : str
        String prefixing all elements of the returned sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Return a generator-based shallow copy of this sequence.
    return get_items_satisfying(
        sequence=sequence,
        item_satisfier=lambda item: (
            types.is_str(item) and item.startswith(item_prefix)))

# ....................{ OMITTERS                          }....................
def omit_item(sequence: SequenceTypes, item: object) -> SequenceTypes:
    '''
    New non-string sequence containing all elements of the first passed
    non-string sequence *not* equalling the passed object.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does *not* modify this sequence.
    item : object
        Object to be omitted from the returned sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    return omit_items(sequence=sequence, items=(item,))


@type_check
def omit_items(
    sequence: SequenceTypes, items: ContainerTypes) -> SequenceTypes:
    '''
    New non-string sequence containing all items of the passed non-string
    sequence *not* contained in the passed non-string container.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to return a proper subset of. For safety, this
        function does *not* modify this sequence.
    items : ContainerTypes
        Container of all items to be omitted from the returned sequence.

    Returns
    ----------
    SequenceTypes
        Proper subset of the passed sequence. For efficiency, this new sequence
        is only a shallow rather than deep copy of the passed sequence.
    '''

    # Return a generator-based shallow copy of this sequence.
    return get_items_satisfying(
        sequence=sequence,
        item_satisfier=lambda item: item not in items,
    )

# ....................{ REMOVERS                          }....................
def remove_item(sequence: SequenceTypes, item: object) -> None:
    '''
    Remove all items equal to the passed object from the passed mutable
    non-string sequence in-place.

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to remove items from in-place.
    item : object
        Object to be removed.
    '''

    remove_items(sequence=sequence, items=(item,))


@type_check
def remove_items(sequence: SequenceTypes, items: ContainerTypes) -> None:
    '''
    Remove all items contained in the passed non-string container from the
    passed mutable non-string sequence in-place.

    Parameters
    ----------
    sequence : SequenceTypes
        Sequence to remove items from in-place.
    items : ContainerTypes
        Container of all items to be removed from this sequence.
    '''

    # Slice assignment implicitly modifies the original sequence, permitting
    # efficient reuse of existing assignment-based functionality.
    sequence[:] = omit_items(sequence, items)

# ....................{ REPLACERS                         }....................
@type_check
def replace_items(
    sequence: SequenceTypes, replacements: MappingType) -> SequenceTypes:
    '''
    New non-string sequence containing all items of the passed non-string
    sequence (_in the same order_) such that each item equal to a key of the
    passed dictionary is replaced by the value of that key.

    This method effectively performs an equality-based global
    search-and-replacement for sequence items.

    Parameters
    ----------
    sequence : SequenceTypes
        Original sequence to be returned transformed. For safety, this
        function does *not* modify this sequence.
    replacements : MappingType
        Mapping whose:

        * Keys are input values to find in the passed sequence. For simplicity,
          keys are matched via object equality rather than more complex object
          matching alternatives (e.g., glob, regex, prefix, suffix, substring).
        * Values are output values to replace these input values with in the
          returned sequence.

    Returns
    ----------
    SequenceTypes
        Sequence transformed from the passed sequence. For efficiency, this
        sequence is only a shallow rather than deep copy of the passed
        sequence.
    '''

    # Type of both the passed sequence and the sequence to be returned.
    sequence_type = type(sequence)

    # Return a shallow copy of this sequence, replacing each item that is a key
    # of the passed mapping by that key's value.
    return sequence_type(
        replacements[item] if item in replacements else item
        for item in sequence
    )

# ....................{ CONVERTERS                        }....................
@type_check
def to_sequence(iterable: IterableTypes) -> SequenceTypes:
    '''
    Sequence converted from the passed iterable if this iterable is not a
    sequence *or* this iterable as is otherwise (i.e., if this iterable is a
    sequence).

    For efficiency, this function only performs a shallow copy of the items in
    this iterable on creating this sequence. Ergo, this sequence is guaranteed
    to contain the same items in the same order as this iterable.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be converted into a sequence.

    Returns
    ----------
    SequenceTypes
        Either:

        * If this iterable is already a sequence, this sequence as is.
        * Else, a new sequence converted from this iterable (and hence
          containing the same items in the same order as this iterable).
    '''

    # One-liners to rule them all.
    return iterable if is_sequence(iterable) else tuple(iterable)
